-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THConversionSystem = {}
local THConversionSystem_mt = THUtils.createClass(THConversionSystem, THAbstractSystem)
local debugFlagId = THDebugUtil.createDebugFlag("THConversionSystem")
THConversionSystem.debugFlagId = debugFlagId
THConversionSystem.DATA_KEY = "conversionSystem"
THConversionSystem.MODIFIER_SEND_BITS = 16
THConversionSystem.CONVERTER = {
    DEFAULT = "THAA_CONVERSION_SYSTEM"
}
function THConversionSystem.registerFruitTypeConversionConditionPaths(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#state", "Required growth state id or value")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#minState", "Minimum required growth state id or value")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#maxState", "Maximum required growth state id or value")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#minMoisture", "Minimum input moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#maxMoisture", "Maximum input moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#converterNames", "FruitTypeConverter names")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#fillTypes", "Original input fillType names")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, "#forcedState", "Force harvesting as specified state")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, xmlPath, "#isPickup", "Input fillTypes are pickup (tipToGround) fillTypes")
    local includeStatesPath = string.format("%s.includeStates.state(?)", xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, includeStatesPath, "#min", "Minimum state to include")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, includeStatesPath, "#max", "Maximum state to include")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, includeStatesPath, "#value", "State to include")
end
function THConversionSystem.registerFruitTypeConversionValuePaths(xmlSchema, xmlPath)
    local toFruitTypePath = xmlPath .. ".toFruitType"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toFruitTypePath, "#name", "Name of new fruitType")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toFruitTypePath, "#state", "New growth state after harvesting")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toFruitTypePath, "#groundType", "New groundType texture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, toFruitTypePath, "#removeHaulm", "Remove haulm layer (if any)")
    local toFillTypePath = xmlPath .. ".toFillType"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toFillTypePath, "#name", "Name of new fillType")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toFillTypePath, "#yield", "Absolute fillType yield (literPerSqm)")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toFillTypePath, "#factor", "Output fillType factor")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toFillTypePath, "#moisture", "Force harvest moisture")
    local toStrawPath = xmlPath .. ".toStraw"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toStrawPath, "#fillType", "New straw fillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toStrawPath, "#yield", "Absolute straw fillType yield (literPerSqm)")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toStrawPath, "#factor", "Straw fillType factor")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, toStrawPath, "#useCutterRatio", "Use header/cutter strawRatio")
    local toChopperPath = xmlPath .. ".toChopper"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toChopperPath, "#fillType", "New chopper fillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toChopperPath, "#yield", "Chopper fillType factor")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, toChopperPath, "#factor", "Chopper fillType factor")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, toChopperPath, "#groundType", "New chopper ground texture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, toChopperPath, "#dropWindrow", "Drop chopped straw material")
end
function THConversionSystem.initXMLSchema(xmlSchema, xmlPath)
    THConversionSystem:superClass().initXMLSchema(xmlSchema, xmlPath)
    local function registerFruitTypeConversions(pBasePath)
        THConversionSystem.registerFruitTypeConversionConditionPaths(xmlSchema, pBasePath)
        THConversionSystem.registerFruitTypeConversionValuePaths(xmlSchema, pBasePath)
    end
    local fruitTypePath = xmlPath .. ".fruitTypes.fruitType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fruitTypePath, "#name", "FruitType name")
    registerFruitTypeConversions(fruitTypePath .. ".conversions.convert(?)")
    local destructionPath = fruitTypePath .. ".destruction.state(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, destructionPath, "#min", "Min conversion growth state")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, destructionPath, "#max", "Max conversion growth state")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, destructionPath, "#to", "Converted growth state")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, destructionPath, "#limitToField", "Limit destruction to field", false)
end
function THConversionSystem.initXMLSchemaSettings(xmlSchema, xmlPath)
    THConversionSystem:superClass().initXMLSchemaSettings(xmlSchema, xmlPath)
end
function THConversionSystem.initXMLSchemaSavegame(xmlSchema, xmlPath)
    THConversionSystem:superClass().initXMLSchemaSavegame(xmlSchema, xmlPath)
end
function THConversionSystem.new(owner, dataKey, modName, customMt)
    dataKey = dataKey or THConversionSystem.DATA_KEY
    customMt = customMt or THConversionSystem_mt
    local self = THAbstractSystem.new(owner, dataKey, modName, customMt)
    self.thSettingsManager = owner.thSettingsManager
    self.thSpecManager = owner.thSpecManager
    self.thFruitTypeManager = owner.thFruitTypeManager
    self.thFillTypeManager = owner.thFillTypeManager
    self.thPFManager = owner.thPFManager
    self.fruitTypes = {}
    return self
end
function THConversionSystem.newModule(owner, modData)
    local self = THConversionSystem.new(owner, nil, modData.name)
    _G.g_thConversionSystem = self
    local modName = self.modName
    local modPath = self.modPath
    local xmlBaseKey = self.xmlBaseKey
    g_xmlManager:addCreateSchemaFunction(function()
        self.xmlSchema = XMLSchema.new(modName)
        self.xmlSchemaSettings = XMLSchema.new(modName .. "_settings")
        self.xmlSchemaSavegame = XMLSchema.new(modName .. "_savegame")
    end)
    g_xmlManager:addInitSchemaFunction(function()
        THConversionSystem.initXMLSchema(self.xmlSchema, xmlBaseKey)
        THConversionSystem.initXMLSchemaSettings(self.xmlSchemaSettings, xmlBaseKey)
        THConversionSystem.initXMLSchemaSavegame(self.xmlSchemaSavegame, xmlBaseKey)
    end)
    local thSpecManager = self.thSpecManager
    source(modPath .. "scripts/utilities/THDensityMapConversionUtil.lua")
    source(modPath .. "scripts/utilities/THDensityMapHeightConversionUtil.lua")
    thSpecManager:addSpecialization("vehicle", "conversionWheels", "THConversionWheels", modName, "scripts/vehicles/specializations/THConversionWheels.lua")
    thSpecManager:addSpecialization("vehicle", "conversionCutter", "THConversionCutter", modName, "scripts/vehicles/specializations/THConversionCutter.lua")
    thSpecManager:addSpecialization("vehicle", "conversionMower", "THConversionMower", modName, "scripts/vehicles/specializations/THConversionMower.lua")
    thSpecManager:addSpecialization("vehicle", "conversionCombine", "THConversionCombine", modName, "scripts/vehicles/specializations/THConversionCombine.lua")
    return self
end
function THConversionSystem:initializeDataTables()
    if self:initFruitTypeData() then
        return true
    end
    return false
end
function THConversionSystem:initFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypesArray, numFruitTypes = thFruitTypeManager:getFruitTypeList(true)
    local defaultConverterName = THConversionSystem.CONVERTER.DEFAULT
    local defaultConverter = thFruitTypeManager:getFruitTypeConverterData(defaultConverterName)
    if fruitTypesArray ~= nil and numFruitTypes > 0 then
        for fruitTypeIndex = 1, numFruitTypes do
            local fruitTypeInfo = fruitTypesArray[fruitTypeIndex]
            local fruitTypeData = thFruitTypeManager:getFruitTypeData(fruitTypeIndex)
            local fruitTypeConversionData = {
                modifiers = {},
                destructionFilters = {},
                pickupFillTypes = {},
                pickupFillTypeMapping = {},
                isEnabled = true
            }
            fruitTypeConversionData.fruitType = {
                index = fruitTypeInfo.index,
                name = fruitTypeInfo.name,
                info = fruitTypeInfo,
                data = fruitTypeData
            }
            if fruitTypeData == nil then
                THUtils.stackMsg("FruitType (%s) data table not found", fruitTypeInfo.name)
                fruitTypeConversionData.isEnabled = false
            elseif defaultConverter == nil or defaultConverter[fruitTypeIndex] == nil then
                local converterData, converterEntry = thFruitTypeManager:addFruitTypeConversion(defaultConverterName, false, fruitTypeIndex, fruitTypeData.fillTypeIndex, 1, 1)
                if converterData == nil or converterEntry == nil then
                    THUtils.stackMsg("Could not add fruitType (%s) to default fruitTypeConverter table", fruitTypeInfo.name)
                    fruitTypeConversionData.isEnabled = false
                elseif defaultConverter == nil then
                    defaultConverter = converterData
                end
            end
            self.fruitTypes[fruitTypeIndex] = fruitTypeConversionData
        end
    end
    return true
end
function THConversionSystem:finalizeDataTables()
    if self:finalizeFruitTypeData() then
        if THDebugUtil.getIsEnabled(debugFlagId) then
            local isSpacePrinted = false
            for _, fruitTypeConversionData in pairs(self.fruitTypes) do
                local numModifiers = #fruitTypeConversionData.modifiers
                if numModifiers > 0 then
                    local fruitTypeIndex = fruitTypeConversionData.fruitType.index
                    local fruitTypeName = fruitTypeConversionData.fruitType.name
                    if not isSpacePrinted then
                        THDebugUtil.printMsg(debugFlagId, nil, "")
                        isSpacePrinted = true
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "Conversion information for fruitType [%s] (%s):", fruitTypeIndex, fruitTypeName)
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                    if fruitTypeConversionData.pickupFillTypes ~= nil
                        and next(fruitTypeConversionData.pickupFillTypes) ~= nil
                    then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Pickup fillTypes:")
                        THDebugUtil.printTable(debugFlagId, nil, fruitTypeConversionData.pickupFillTypes, 1)
                    end
                    for modifierIndex = 1, numModifiers do
                        local modifierData = fruitTypeConversionData.modifiers[modifierIndex]
                        THDebugUtil.printMsg(debugFlagId, nil, "- Modifier #%0d:", modifierIndex)
                        THDebugUtil.printTable(debugFlagId, nil, modifierData.values)
                        THDebugUtil.printMsg(debugFlagId, nil, "- Conditions:")
                        THDebugUtil.printTable(debugFlagId, nil, modifierData.conditions, 1)
                    end
                end
            end
        end
        return true
    end
    return false
end
function THConversionSystem:finalizeFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local _, numFruitTypes = thFruitTypeManager:getFruitTypeList(true)
    for fruitTypeIndex = 1, numFruitTypes do
        self:updateFruitTypePickupFillTypes(fruitTypeIndex)
    end
    return true
end
function THConversionSystem:initSettings()
    local thSettingsManager = self.thSettingsManager
    local categoryMain = thSettingsManager:addSettingCategory(self.dataKey, nil, self.modName)
    local settingUseChoppedStraw = thSettingsManager:addSetting("useChoppedStraw", categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.CHECKED, self.modName, false, true, true)
    settingUseChoppedStraw:setValue(true, true, true)
    thSettingsManager:addSetting("aiCropDestruction", categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.CHECKED, self.modName, false, true, true)
    local settingTipDestruction = thSettingsManager:addSetting("tipDestruction", categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    settingTipDestruction:addValue("OFF", "$l10n_ui_off")
    settingTipDestruction:addValue("FULL", "$l10n_thUI_full")
    settingTipDestruction:addValue("CUTPLUS", "$l10n_thUI_cutPlus")
    settingTipDestruction:addValue("CUTONLY", "$l10n_thUI_cutOnly")
    settingTipDestruction:setValue("OFF", true, true)
end
function THConversionSystem:onPreInit()
    if not self.isPreInitFinished then
        THConversionSystem:superClass().onPreInit(self)
        self:initSettings()
        self:addDelayedLoadTask(function()
            THUtils.setFunctionHook("FruitTypeManager", "loadDefaultTypes", THConversionSystem, THConversionSystem.hook_loadDefaultFruitTypes, false, self)
            THUtils.setFunctionHook("FruitTypeManager", "loadMapData", THConversionSystem, THConversionSystem.hook_loadFruitTypeMapData, false, self)
        end)
        self.isPreInitFinished = true
    end
end
function THConversionSystem:onPostInit()
    if not self.isPostInitFinished then
        THConversionSystem:superClass().onPostInit(self)
        self.isPostInitFinished = true
    end
end
function THConversionSystem:onLoadEnvironment(environment, mission)
    THConversionSystem:superClass().onLoadEnvironment(self, environment, mission)
end
function THConversionSystem:onLoadMapFinished(mission)
    THConversionSystem:superClass().onLoadMapFinished(self, mission)
    self:initializeDataTables()
    self:loadConfiguration()
    self:finalizeDataTables()
end
function THConversionSystem:loadFromConfigXMLFile(xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local loadSuccess = THConversionSystem:superClass().loadFromConfigXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        local fruitTypesKey = xmlKey .. ".fruitTypes"
        xmlFile:iterate(fruitTypesKey .. ".fruitType", function(_, pFruitTypeKey)
            local fruitTypeName = THUtils.getXMLValue(xmlFile, pFruitTypeKey, "#name")
            local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeName)
            if fruitTypeInfo == nil or fruitTypeInfo.index == FruitType.UNKNOWN then
                THUtils.xmlLoadMsg(pFruitTypeKey, "name", fruitTypeName, fruitTypeName == nil, THXMLMessage.INVALID_FRUIT_TYPE)
            else
                if not self:loadFruitTypeConversionData(fruitTypeInfo.index, xmlFile, pFruitTypeKey)
                    or not self:loadFruitTypeDestructionData(fruitTypeInfo.index, xmlFile, pFruitTypeKey)
                then
                    loadSuccess = false
                    return false
                end
            end
        end)
    end
    return loadSuccess
end
function THConversionSystem:loadFruitTypeConversionData(fruitType, xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local thFillTypeManager = self.thFillTypeManager
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    local loadSuccess = true
    if fruitTypeConversionData ~= nil then
        local fruitTypeData = fruitTypeConversionData.fruitType.data
        local fruitTypeIndex = fruitTypeConversionData.fruitType.index
        local numGrowthStates = fruitTypeData.numGrowthStates or 0
        local areModifiersReset = false
        if numGrowthStates > 0 then
            xmlFile:iterate(xmlKey .. ".conversions.convert", function(_, pConversionKey)
                local growthStateId = THUtils.getXMLValue(xmlFile, pConversionKey, "#state")
                local minGrowthStateId = THUtils.getXMLValue(xmlFile, pConversionKey, "#minState")
                local maxGrowthStateId = THUtils.getXMLValue(xmlFile, pConversionKey, "#maxState")
                local converterNames = THUtils.getXMLValue(xmlFile, pConversionKey, "#converterNames")
                local inputFillTypes = THUtils.getXMLValue(xmlFile, pConversionKey, "#fillTypes")
                local minMoisture = THUtils.getXMLValue(xmlFile, pConversionKey, "#minMoisture")
                local maxMoisture = THUtils.getXMLValue(xmlFile, pConversionKey, "#maxMoisture")
                local forcedState = THUtils.getXMLValue(xmlFile, pConversionKey, "#forcedState")
                local isPickup = THUtils.getXMLValue(xmlFile, pConversionKey, "#isPickup")
                local modifierData = {
                    parent = fruitTypeConversionData,
                    fruitType = fruitTypeConversionData.fruitType,
                    values = {},
                    conditions = {
                        allowedStates = {},
                        converterNames = {},
                        inputFillTypes = {}
                    }
                }
                local conditionData = modifierData.conditions
                local areConditionsValid = true
                local function addAllowedStates(pMinState, pMaxState)
                    if pMaxState == nil then
                        conditionData.allowedStates[pMinState] = true
                    else
                        if pMaxState < pMinState then
                            pMaxState = pMinState
                        end
                        for stateIndex = pMinState, pMaxState do
                            conditionData.allowedStates[stateIndex] = true
                        end
                    end
                end
                if growthStateId ~= nil then
                    local growthState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, growthStateId)
                    if growthState == nil or growthState < 0 then
                        THUtils.xmlLoadMsg(pConversionKey, "state", growthStateId, true, "invalid growth state")
                        areConditionsValid = false
                    else
                        addAllowedStates(growthState)
                    end
                elseif minGrowthStateId ~= nil or maxGrowthStateId ~= nil then
                    local minState = 0
                    local maxState = numGrowthStates
                    if minGrowthStateId ~= nil then
                        minState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, minGrowthStateId)
                        if minState == nil or minState < 0 then
                            THUtils.xmlLoadMsg(pConversionKey, "minState", minGrowthStateId, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, numGrowthStates)
                            areConditionsValid = false
                        end
                    end
                    if areConditionsValid and maxGrowthStateId ~= nil then
                        maxState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, maxGrowthStateId)
                        if maxState == nil or maxState < 0 then
                            THUtils.xmlLoadMsg(pConversionKey, "maxState", maxGrowthStateId, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, numGrowthStates)
                            areConditionsValid = false
                        end
                    end
                    if areConditionsValid then
                        addAllowedStates(minState, maxState)
                    end
                end
                if areConditionsValid and converterNames ~= nil then
                    local convertersArray = THUtils.splitString(converterNames, " ")
                    local hasConverterError = false
                    if convertersArray ~= nil and #convertersArray > 0 then
                        for _, converterName in ipairs(convertersArray) do
                            converterName = converterName:upper()
                            if converterName == "NONE" then
                                conditionData.converterNames[converterName] = true
                            elseif converterName ~= "" then
                                local fruitTypeConverter = thFruitTypeManager:getFruitTypeConverterData(converterName)
                                if fruitTypeConverter == nil then
                                    THUtils.xmlLoadMsg(pConversionKey, "converterName", converterNames, false, "fruitTypeConverter not found")
                                    hasConverterError = true
                                else
                                    conditionData.converterNames[converterName] = true
                                end
                            end
                        end
                    end
                    if next(conditionData.converterNames) == nil then
                        if not hasConverterError then
                            THUtils.xmlLoadMsg(pConversionKey, "converterNames", converterNames, true, "no valid values found")
                        end
                        areConditionsValid = false
                    end
                end
                if areConditionsValid and inputFillTypes ~= nil then
                    local fillTypesArray = THUtils.splitString(inputFillTypes, " ")
                    local hasFillTypeError = false
                    if fillTypesArray ~= nil and #fillTypesArray > 0 then
                        for _, fillTypeName in ipairs(fillTypesArray) do
                            if fillTypeName ~= "" then
                                local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
                                if fillTypeInfo == nil or fillTypeInfo.index == FillType.UNKNOWN then
                                    if not hasFillTypeError then
                                        THUtils.xmlLoadMsg(pConversionKey, "fillType", fillTypeName, false, THXMLMessage.INVALID_FILL_TYPE)
                                        hasFillTypeError = true
                                    end
                                else
                                    conditionData.inputFillTypes[fillTypeInfo.index] = true
                                end
                            end
                        end
                    end
                    if next(conditionData.inputFillTypes) == nil then
                        if not hasFillTypeError then
                            THUtils.xmlLoadMsg(pConversionKey, "fillTypes", inputFillTypes, true, "no valid values found")
                        end
                        areConditionsValid = false
                    end
                end
                if areConditionsValid and (minMoisture ~= nil or maxMoisture ~= nil) then
                    local realMinMoisture = 0
                    local realMaxMoisture = 1
                    if minMoisture ~= nil then
                        if minMoisture < 0 or minMoisture > 100 then
                            THUtils.xmlLoadMsg(pConversionKey, "minMoisture", minMoisture, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                            areConditionsValid = false
                        else
                            realMinMoisture = minMoisture / 100
                        end
                    end
                    if areConditionsValid and maxMoisture ~= nil then
                        if maxMoisture < 0 or maxMoisture > 100 then
                            THUtils.xmlLoadMsg(pConversionKey, "maxMoisture", maxMoisture, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                            areConditionsValid = false
                        else
                            realMaxMoisture = maxMoisture / 100
                        end
                    end
                    if areConditionsValid then
                        conditionData.minMoisture = math.min(realMinMoisture, realMaxMoisture)
                        conditionData.maxMoisture = math.max(realMinMoisture, realMaxMoisture)
                    end
                end
                if areConditionsValid and forcedState ~= nil then
                    forcedState = forcedState:upper()
                    if forcedState == "FORAGE" then
                        if fruitTypeData.minForageState ~= nil and fruitTypeData.minForageState > 0 then
                            conditionData.forcedState = fruitTypeData.minForageState
                        else
                            THUtils.xmlLoadMsg(pConversionKey, "forcedState", forcedState, false, "growth state does not exist")
                            areConditionsValid = false
                        end
                    elseif forcedState == "HARVEST" then
                        if fruitTypeData.minHarvestState ~= nil and fruitTypeData.minHarvestState > 0 then
                            conditionData.forcedState = fruitTypeData.minHarvestState
                        else
                            THUtils.xmlLoadMsg(pConversionKey, "forcedState", forcedState, false, "growth state does not exist")
                            areConditionsValid = false
                        end
                    else
                        THUtils.xmlLoadMsg(pConversionKey, "forcedState", forcedState, true, "must be either FORAGE or HARVEST")
                        areConditionsValid = false
                    end
                end
                if areConditionsValid then
                    conditionData.isPickup = isPickup
                    xmlFile:iterate(pConversionKey .. ".includeStates.state", function(_, pStateKey)
                        local includeStateId = THUtils.getXMLValue(xmlFile, pStateKey, "#value")
                        local includeMinStateId = THUtils.getXMLValue(xmlFile, pStateKey, "#min")
                        local includeMaxStateId = THUtils.getXMLValue(xmlFile, pStateKey, "#max")
                        if includeStateId ~= nil then
                            local includeState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, includeStateId)
                            if includeState == nil or includeState < 0 then
                                THUtils.xmlLoadMsg(pStateKey, "value", includeStateId, true, THXMLMessage.INVALID_VALUE)
                            else
                                addAllowedStates(includeState)
                            end
                        elseif includeMinStateId ~= nil or includeMaxStateId ~= nil then
                            local includeMinState = -1
                            local includeMaxState = -1
                            if includeMinStateId ~= nil then
                                includeMinState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, includeMinStateId)
                            end
                            if includeMaxStateId ~= nil then
                                includeMaxState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, includeMaxStateId)
                            end
                            if includeMinState == nil or includeMinState < 0 then
                                THUtils.xmlLoadMsg(pStateKey, "min", includeMinStateId, true, THXMLMessage.INVALID_VALUE)
                            elseif includeMaxState == nil or includeMaxState < 0 then
                                THUtils.xmlLoadMsg(pStateKey, "max", includeMaxStateId, true, THXMLMessage.INVALID_VALUE)
                            else
                                addAllowedStates(includeMinState, includeMaxState)
                            end
                        end
                    end)
                    if self:loadFruitTypeConversionValues(modifierData, xmlFile, pConversionKey) then
                        if not areModifiersReset then
                            THUtils.clearTable(fruitTypeConversionData.modifiers)
                            areModifiersReset = true
                        end
                        local maxNumModifiers = 2 ^ (THConversionSystem.MODIFIER_SEND_BITS) - 1
                        local numModifiers = #fruitTypeConversionData.modifiers
                        if numModifiers < maxNumModifiers then
                            numModifiers = numModifiers + 1
                            modifierData.index = numModifiers
                            fruitTypeConversionData.modifiers[numModifiers] = modifierData
                        else
                            THUtils.warningMsg("Too many modifiers for fruitType [%s] (%s): %s", fruitTypeData.index, fruitTypeData.name, numModifiers)
                        end
                    end
                end
            end)
        end
    end
    return loadSuccess
end
function THConversionSystem:loadFruitTypeConversionValues(modifierData, xmlFile, xmlKey)
    local mission = self.mission
    local fieldGroundSystem = mission.fieldGroundSystem
    local fruitTypeData = modifierData.fruitType.data
    local fruitTypeInfo = modifierData.fruitType.info
    local conditionData = modifierData.conditions
    local modifierValues = THUtils.clearTable(modifierData.values)
    local areValuesValid = true
    local toFruitTypeKey = string.format("%s.toFruitType", xmlKey)
    local toFruitTypeName = THUtils.getXMLValue(xmlFile, toFruitTypeKey, "#name")
    local toCutStateId = THUtils.getXMLValue(xmlFile, toFruitTypeKey, "#state")
    local toGroundType = THUtils.getXMLValue(xmlFile, toFruitTypeKey, "#groundType")
    local removeHaulm = THUtils.getXMLValue(xmlFile, toFruitTypeKey, "#removeHaulm")
    local toFillTypeKey = string.format("%s.toFillType", xmlKey)
    local toFillTypeName = THUtils.getXMLValue(xmlFile, toFillTypeKey, "#name")
    local toFillTypeYield = THUtils.getXMLValue(xmlFile, toFillTypeKey, "#yield")
    local toFillTypeFactor = THUtils.getXMLValue(xmlFile, toFillTypeKey, "#factor")
    local toMoisture = THUtils.getXMLValue(xmlFile, toFillTypeKey, "#moisture")
    local toStrawKey = string.format("%s.toStraw", xmlKey)
    local toStrawFillType = THUtils.getXMLValue(xmlFile, toStrawKey, "#fillType")
    local toStrawYield = THUtils.getXMLValue(xmlFile, toStrawKey, "#yield")
    local toStrawFactor = THUtils.getXMLValue(xmlFile, toStrawKey, "#factor")
    local useCutterRatio = THUtils.getXMLValue(xmlFile, toStrawKey, "#useCutterRatio")
    local toChopperKey = string.format("%s.toChopper", xmlKey)
    local toChopperFillType = THUtils.getXMLValue(xmlFile, toChopperKey, "#fillType")
    local toChopperYield = THUtils.getXMLValue(xmlFile, toChopperKey, "#yield")
    local toChopperFactor = THUtils.getXMLValue(xmlFile, toChopperKey, "#factor")
    local toChopperGroundType = THUtils.getXMLValue(xmlFile, toChopperKey, "#groundType")
    local useChoppedStraw = THUtils.getXMLValue(xmlFile, toChopperKey, "#dropWindrow")
    if toFruitTypeName ~= nil then
        local toFruitTypeInfo = self.thFruitTypeManager:getFruitType(toFruitTypeName)
        if toFruitTypeInfo == nil or toFruitTypeInfo.index == FruitType.UNKNOWN then
            THUtils.xmlLoadMsg(toFruitTypeKey, "name", toFruitTypeName, false, THXMLMessage.INVALID_FRUIT_TYPE)
            areValuesValid = false
        else
            modifierValues.toFruitType = toFruitTypeInfo.index
        end
    end
    if areValuesValid and toCutStateId ~= nil then
        local toCutState = nil -- placeholders
        if modifierValues.toFruitType ~= nil then
            toCutState = self.thFruitTypeManager:getFruitTypeGrowthState(modifierValues.toFruitType, toCutStateId, true)
        else
            toCutState = self.thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, toCutStateId, true)
        end
        if toCutState == nil or toCutState < 0 then
            THUtils.xmlLoadMsg(toFruitTypeKey, "state", toCutStateId, true, THXMLMessage.INVALID_VALUE)
            areValuesValid = false
        else
            modifierValues.toCutState = toCutState
        end
    end
    if areValuesValid and toGroundType ~= nil then
        toGroundType = toGroundType:upper()
        local toGroundTypeIndex = FieldGroundType.getByName(toGroundType)
        if toGroundTypeIndex == nil then
            THUtils.xmlLoadMsg(toFruitTypeKey, "groundType", toGroundType, true, "invalid ground type")
            areValuesValid = false
        else
            modifierValues.toGroundType = toGroundType
            modifierValues.toGroundTypeIndex = toGroundTypeIndex
        end
    end
    if areValuesValid and toFillTypeName ~= nil then
        local toFillTypeInfo = self.thFillTypeManager:getFillType(toFillTypeName)
        if toFillTypeInfo == nil or toFillTypeInfo.index == FillType.UNKNOWN then
            THUtils.xmlLoadMsg(toFillTypeKey, "name", toFillTypeName, false, THXMLMessage.INVALID_FILL_TYPE)
            areValuesValid = false
        else
            modifierValues.toFillType = toFillTypeInfo.index
        end
    end
    if areValuesValid and toFillTypeYield ~= nil then
        if toFillTypeYield < 0 then
            THUtils.xmlLoadMsg(toFillTypeKey, "yield", toFillTypeYield, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
            areValuesValid = false
        elseif fruitTypeInfo.literPerSqm == nil or fruitTypeInfo.literPerSqm == 0 then
            THUtils.xmlLoadMsg(toFillTypeKey, "yield", toFillTypeYield, true, "cannot set yield of non-yield fruitType")
            areValuesValid = false
        else
            if toFillTypeFactor ~= nil then
                THUtils.xmlLoadMsg(toFillTypeKey, "yield", toFillTypeYield, false, "yield will override factor (%s) setting", toFillTypeFactor)
            end
            toFillTypeFactor = toFillTypeYield / fruitTypeInfo.literPerSqm
        end
    end
    if areValuesValid and toFillTypeFactor ~= nil then
        if toFillTypeFactor < 0 then
            THUtils.xmlLoadMsg(toFillTypeKey, "factor", toFillTypeFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
            areValuesValid = false
        else
            modifierValues.toFillTypeFactor = toFillTypeFactor
        end
    end
    if areValuesValid and toMoisture ~= nil then
        if toMoisture < 0 or toMoisture > 100 then
            THUtils.xmlLoadMsg(toFillTypeKey, "moisture", toMoisture, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
            areValuesValid = false
        else
            modifierValues.toMoisture = MathUtil.clamp(toMoisture / 100, 0, 1)
        end
    end
    local strawYield = fruitTypeInfo.windrowLiterPerSqm or fruitTypeInfo.literPerSqm
    if areValuesValid
        and (toStrawFillType ~= nil or toStrawFactor ~= nil or toStrawYield ~= nil)
    then
        if toStrawFillType ~= nil then
            local toStrawFillTypeInfo = self.thFillTypeManager:getFillType(toStrawFillType)
            if toStrawFillTypeInfo == nil or toStrawFillTypeInfo.index == FillType.UNKNOWN then
                THUtils.xmlLoadMsg(toStrawKey, "fillType", toStrawFillType, false, THXMLMessage.INVALID_FILL_TYPE)
                areValuesValid = false
            else
                modifierValues.toStrawFillType = toStrawFillTypeInfo.index
            end
        end
        if areValuesValid and toStrawYield ~= nil and toStrawFactor ~= 0 then
            if toStrawYield < 0 then
                THUtils.xmlLoadMsg(toStrawKey, "yield", toStrawYield, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                areValuesValid = false
            elseif strawYield == nil or strawYield <= 0 then
                THUtils.xmlLoadMsg(toStrawKey, "yield", toStrawYield, true, "cannot set yield of non-yield fruitType")
                areValuesValid = false
            else
                if toStrawFactor ~= nil then
                    THUtils.xmlLoadMsg(toStrawKey, "yield", toStrawYield, false, "yield will override factor (%s) setting", toStrawFactor)
                end
                toStrawFactor = toStrawYield / strawYield
            end
        end
        if areValuesValid and toStrawFactor ~= nil then
            if toStrawFactor < 0 then
                THUtils.xmlLoadMsg(toStrawKey, "factor", toStrawFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                areValuesValid = false
            else
                modifierValues.toStrawFactor = toStrawFactor
            end
        end
    end
    if areValuesValid
        and (toChopperFillType ~= nil or toChopperYield ~= nil or toChopperFactor ~= nil)
    then
        if toChopperFillType ~= nil then
            local toChopperFillTypeInfo = self.thFillTypeManager:getFillType(toChopperFillType)
            if toChopperFillTypeInfo == nil or toChopperFillTypeInfo.index == FillType.UNKNOWN then
                THUtils.xmlLoadMsg(toChopperKey, "fillType", toChopperFillType, false, THXMLMessage.INVALID_FILL_TYPE)
                areValuesValid = false
            else
                modifierValues.toChopperFillType = toChopperFillTypeInfo.index
            end
        end
        if areValuesValid and toChopperYield ~= nil then
            if toChopperYield < 0 then
                THUtils.xmlLoadMsg(toChopperKey, "yield", toChopperYield, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                areValuesValid = false
            elseif strawYield == nil or strawYield <= 0 then
                THUtils.xmlLoadMsg(toChopperKey, "yield", toChopperYield, true, "cannot set yield of non-yield fruitType")
                areValuesValid = false
            else
                if toChopperFactor ~= nil then
                    THUtils.xmlLoadMsg(toChopperKey, "yield", toChopperYield, false, "yield will override factor (%s) setting", toChopperFactor)
                end
                toChopperFactor = toChopperYield / strawYield
            end
        end
        if areValuesValid and toChopperFactor ~= nil then
            if toChopperFactor < 0 then
                THUtils.xmlLoadMsg(toChopperKey, "factor", toChopperFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                areValuesValid = false
            else
                modifierValues.toChopperFactor = toChopperFactor
            end
        end
    end
    if areValuesValid and toChopperGroundType ~= nil then
        if fruitTypeInfo.chopperTypeIndex == nil then
            THUtils.xmlLoadMsg(toChopperKey, "groundType", toChopperGroundType, true, "fruitType (%s) does not have a chopper groundType to convert", fruitTypeInfo.name)
            areValuesValid = false
        else
            toChopperGroundType = toChopperGroundType:upper()
            if toChopperGroundType == "NONE" then
                modifierValues.toChopperType = toChopperGroundType
                modifierValues.toChopperValue = nil
            else
                local chopperGroundTypeId = FieldChopperType[toChopperGroundType]
                if chopperGroundTypeId == nil then
                    THUtils.xmlLoadMsg(toChopperKey, "groundType", toChopperGroundType, true, "invalid chopper groundType")
                    areValuesValid = false
                else
                    local toChopperValue = fieldGroundSystem:getChopperTypeValue(chopperGroundTypeId)
                    if toChopperValue == nil then
                        THUtils.xmlLoadMsg(toChopperKey, "groundType", toChopperGroundType, true, "invalid chopper groundType")
                        areValuesValid = false
                    else
                        modifierValues.toChopperType = chopperGroundTypeId
                        modifierValues.toChopperValue = toChopperValue
                    end
                end
            end
        end
    end
    if areValuesValid then
        modifierValues.toHarvestState = conditionData.forcedState
        modifierValues.useCutterRatio = useCutterRatio
        modifierValues.removeHaulm = removeHaulm
        modifierValues.useChoppedStraw = useChoppedStraw
        if next(modifierValues) ~= nil then
            return true
        end
        THUtils.xmlLoadMsg(xmlKey, nil, nil, true, "Conversion has no valid attributes")
    end
    return false
end
function THConversionSystem:loadFruitTypeDestructionData(fruitType, xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    local loadSuccess = true
    if fruitTypeConversionData ~= nil then
        local fruitTypeData = fruitTypeConversionData.fruitType.data
        local maxNumStates = fruitTypeData.numGrowthStates
        local destructionKey = string.format("%s.destruction", xmlKey)
        if xmlFile:hasProperty(destructionKey) then
            local destructionFilters = THUtils.clearTable(fruitTypeConversionData.destructionFilters)
            local defaultDestruction = fruitTypeData.info.destruction
            if defaultDestruction == nil then
                THUtils.warningMsg("Cannot assign destruction data to fruitType (%s), no default destruction data exists", fruitTypeData.name)
            else
                local areFiltersReset = false
                local isDefaultAdded = false
                xmlFile:iterate(xmlKey .. ".destruction.state", function(_, pDestructionKey)
                    local minStateId = THUtils.getXMLValue(xmlFile, pDestructionKey, "#min")
                    local maxStateId = THUtils.getXMLValue(xmlFile, pDestructionKey, "#max")
                    local toStateId = THUtils.getXMLValue(xmlFile, pDestructionKey, "#to")
                    local limitToField = THUtils.getXMLValue(xmlFile, pDestructionKey, "#limitToField", false)
                    local minState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, minStateId, true)
                    local maxState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, maxStateId, true)
                    local toState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, toStateId, true)
                    if minState == nil or minState < 0 or minState > maxNumStates then
                        THUtils.xmlLoadMsg(pDestructionKey, "min", minStateId, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, maxNumStates)
                    elseif maxState == nil or maxState < 0 or maxState > maxNumStates then
                        THUtils.xmlLoadMsg(pDestructionKey, "max", maxStateId, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, maxNumStates)
                    elseif toStateId == nil or toState < 0 or toState > maxNumStates then
                        THUtils.xmlLoadMsg(pDestructionKey, "to", toStateId, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, maxNumStates)
                    else
                        if maxState < minState then
                            maxState = minState
                        end
                        local destructionData = {
                            minState = minState,
                            maxState = maxState,
                            toState = toState,
                            limitToField = limitToField
                        }
                        if not areFiltersReset then
                            THUtils.clearTable(destructionFilters)
                            areFiltersReset = true
                        end
                        if not isDefaultAdded then
                            defaultDestruction.canBeDestroyed = true
                            defaultDestruction.filterStart = minState
                            defaultDestruction.filterEnd = maxState
                            defaultDestruction.state = toState
                            defaultDestruction.onlyOnField = limitToField == true
                            isDefaultAdded = true
                        else
                            table.insert(destructionFilters, destructionData)
                        end
                    end
                end)
            end
        end
    end
    return loadSuccess
end
function THConversionSystem:getFruitTypeConversionData(fruitType, force)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
    if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= nil then
        local fruitTypeConversionData = self.fruitTypes[fruitTypeInfo.index]
        if fruitTypeConversionData ~= nil and (fruitTypeConversionData.isEnabled or force == true) then
            return fruitTypeConversionData
        end
    end
end
function THConversionSystem:getFruitTypePickupData(fruitType, fillType, growthStateId)
    local growthState = nil
    if growthStateId ~= nil then
        growthState = self.thFruitTypeManager:getFruitTypeGrowthState(fruitType, growthStateId, true)
        if growthState == nil or growthState < 0 then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "growthStateId", growthStateId)
            return
        end
    end
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    if fruitTypeConversionData ~= nil then
        local numPickupFillTypes = #fruitTypeConversionData.pickupFillTypes
        if numPickupFillTypes > 0 then
            if fillType ~= nil then
                local fillTypeInfo = self.thFillTypeManager:getFillType(fillType)
                if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                    local pickupIndex = fruitTypeConversionData.pickupFillTypeMapping[fillTypeInfo.index]
                    local pickupData = fruitTypeConversionData.pickupFillTypes[pickupIndex]
                    if pickupData ~= nil then
                        if growthState == nil
                            or (growthState >= pickupData.minState and growthState <= pickupData.maxState)
                        then
                            return pickupData
                        end
                    end
                end
            else
                for pickupIndex = 1, numPickupFillTypes do
                    local pickupData = fruitTypeConversionData.pickupFillTypes[pickupIndex]
                    if growthState == nil
                        or (growthState >= pickupData.minState and growthState <= pickupData.maxState)
                    then
                        return pickupData
                    end
                end
            end
        end
    end
end
function THConversionSystem:updateFruitTypePickupFillTypes(fruitType)
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    if fruitTypeConversionData ~= nil then
        THUtils.clearTable(fruitTypeConversionData.pickupFillTypes)
        THUtils.clearTable(fruitTypeConversionData.pickupFillTypeMapping)
        local fruitTypeData = fruitTypeConversionData.fruitType.data
        local fruitTypeIndex = fruitTypeConversionData.fruitType.index
        local numGrowthStates = fruitTypeData.numGrowthStates or 0
        if numGrowthStates > 0 then
            local function addPickupFillType(pFillTypeIndex, pMinState, pMaxState)
                local pickupDataIndex = fruitTypeConversionData.pickupFillTypeMapping[pFillTypeIndex]
                local pickupData = fruitTypeConversionData.pickupFillTypes[pickupDataIndex]
                if pickupData == nil then
                    pickupData = {
                        fruitTypeIndex = fruitTypeIndex,
                        fillTypeIndex = pFillTypeIndex,
                        minState = pMinState,
                        maxState = pMaxState,
                    }
                    table.insert(fruitTypeConversionData.pickupFillTypes, pickupData)
                    fruitTypeConversionData.pickupFillTypeMapping[pFillTypeIndex] = #fruitTypeConversionData.pickupFillTypes
                else
                    pickupData.minState = math.min(pickupData.minState, pMinState)
                    pickupData.maxState = math.max(pickupData.maxState, pMaxState)
                end
                pickupData.isWindrow = fruitTypeData.windrowIndex ~= nil and fruitTypeData.windrowIndex ~= FruitType.UNKNOWN
                    and pFillTypeIndex == fruitTypeData.windrowIndex
            end
            if fruitTypeData.windrowIndex ~= nil and fruitTypeData.windrowIndex ~= FillType.UNKNOWN
                and fruitTypeData.minHarvestState ~= nil and fruitTypeData.minHarvestState > 0
                and fruitTypeData.maxHarvestState ~= nil and fruitTypeData.maxHarvestState > 0
            then
                addPickupFillType(fruitTypeData.windrowIndex, fruitTypeData.minHarvestState, fruitTypeData.maxHarvestState)
            end
            local numModifiers = #fruitTypeConversionData.modifiers
            if numModifiers > 0 then
                for modifierIndex = 1, numModifiers do
                    local modifierData = fruitTypeConversionData.modifiers[modifierIndex]
                    local conditionData = modifierData.conditions
                    if conditionData.isPickup and conditionData.inputFillTypes ~= nil then
                        local minState = nil
                        local maxState = nil
                        for stateIndex in pairs(conditionData.allowedStates) do
                            if minState ~= nil then
                                minState = math.min(minState, stateIndex)
                            else
                                minState = stateIndex
                            end
                            if maxState ~= nil then
                                maxState = math.max(maxState, stateIndex)
                            else
                                maxState = stateIndex
                            end
                        end
                        if minState ~= nil and maxState ~= nil then
                            for inputFillType in pairs(conditionData.inputFillTypes) do
                                addPickupFillType(inputFillType, minState, maxState)
                            end
                        end
                    end
                end
            end
        end
    end
end
function THConversionSystem:getFruitTypeConversionFillTypes(fruitType, growthStateId, converterName, inputFillType, isPickup, valuesTable)
    local thFillTypeManager = self.thFillTypeManager
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    local growthState = nil
    if growthStateId ~= nil then
        growthState = self.thFruitTypeManager:getFruitTypeGrowthState(fruitType, growthStateId, true)
        if growthState == nil or growthState < 0 then
            return
        end
    end
    if converterName ~= nil then
        converterName = converterName:upper()
    end
    if inputFillType ~= nil then
        local fillTypeInfo = thFillTypeManager:getFillType(inputFillType)
        if fillTypeInfo == nil then
            return
        end
        inputFillType = fillTypeInfo.index
    end
    if fruitTypeConversionData ~= nil then
        local numModifiers = #fruitTypeConversionData.modifiers
        if numModifiers > 0 then
            for modifierIndex = 1, numModifiers do
                local modifierData = fruitTypeConversionData.modifiers[modifierIndex]
                local modifierValues = modifierData.values
                local conditionData = modifierData.conditions
                if modifierValues.toFillType ~= nil then
                    local isModifierValid = true
                    if growthState ~= nil then
                        if not conditionData.allowedStates[growthState] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and isPickup ~= nil then
                        if (isPickup == true and not conditionData.isPickup)
                            or (isPickup == false and conditionData.isPickup)
                        then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and converterName ~= nil
                        and next(conditionData.converterNames) ~= nil
                    then
                        if not conditionData.converterNames[converterName] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and inputFillType ~= nil
                        and next(conditionData.inputFillTypes) ~= nil
                    then
                        if not conditionData.inputFillTypes[inputFillType] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid then
                        valuesTable = THUtils.getNoNil(valuesTable, {})
                        valuesTable[modifierValues.toFillType] = true
                    end
                end
            end
        end
    end
    return valuesTable
end
function THConversionSystem:getFruitTypeModifierData(fruitType, growthStateId, converterName, inputFillType, inputMoisture, isPickup, supportedFillTypes)
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    if fruitTypeConversionData ~= nil
        and THUtils.argIsValid(converterName == nil or type(converterName) == "string", "converterName", converterName)
    then
        local fruitTypeData = fruitTypeConversionData.fruitType.data
        local growthState = self.thFruitTypeManager:getFruitTypeGrowthState(fruitTypeData.index, growthStateId, true)
        inputMoisture = inputMoisture or -1
        if converterName == nil then
            converterName = "NONE"
        else
            converterName = converterName:upper()
        end
        if inputFillType == nil then
            if converterName == "NONE" then
                inputFillType = fruitTypeData.fillTypeIndex
            else
                local fruitTypeConverter = self.thFruitTypeManager:getFruitTypeConverterData(converterName)
                if fruitTypeConverter ~= nil then
                    local converterInfo = fruitTypeConverter[fruitTypeData.index]
                    if converterInfo ~= nil then
                        inputFillType = converterInfo.fillTypeIndex
                    end
                end
            end
        end
        local inputFillTypeInfo = self.thFillTypeManager:getFillType(inputFillType)
        if THUtils.argIsValid(inputFillTypeInfo ~= nil, "inputFillType", inputFillType)
            and THUtils.argIsValid(growthState ~= nil, "growthState", growthState)
            and THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture)
        then
            inputFillType = inputFillTypeInfo.index
            local numModifiers = #fruitTypeConversionData.modifiers
            if numModifiers > 0 then
                for modifierIndex = 1, numModifiers do
                    local modifierData = fruitTypeConversionData.modifiers[modifierIndex]
                    local modifierValues = modifierData.values
                    local conditionData = modifierData.conditions
                    local isModifierValid = true
                    if supportedFillTypes ~= nil then
                        if modifierValues.toFillType ~= nil then
                            if not supportedFillTypes[modifierValues.toFillType] then
                                isModifierValid = false
                            end
                        else
                            if not supportedFillTypes[inputFillType] then
                                isModifierValid = false
                            end
                        end
                    end
                    if isModifierValid then
                        if not conditionData.allowedStates[growthState] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and isPickup ~= nil then
                        if (isPickup == true and not conditionData.isPickup)
                            or (not isPickup and conditionData.isPickup)
                        then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and conditionData.minMoisture ~= nil then
                        if inputMoisture < 0 or inputMoisture < conditionData.minMoisture then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and conditionData.maxMoisture ~= nil then
                        if inputMoisture < 0 or inputMoisture > conditionData.maxMoisture then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and next(conditionData.converterNames) ~= nil then
                        if not conditionData.converterNames[converterName] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid and next(conditionData.inputFillTypes) ~= nil then
                        if not conditionData.inputFillTypes[inputFillType] then
                            isModifierValid = false
                        end
                    end
                    if isModifierValid then
                        return modifierData
                    end
                end
            end
        end
    end
end
function THConversionSystem:getFruitTypeModifierDataByIndex(fruitType, modifierIndex)
    local fruitTypeConversionData = self:getFruitTypeConversionData(fruitType)
    if fruitTypeConversionData ~= nil then
        if modifierIndex ~= nil then
            return fruitTypeConversionData.modifiers[modifierIndex]
        end
    end
end
function THConversionSystem.hook_loadFruitTypeMapData(superFunc, self, manager, xmlFile, missionInfo, baseDirectory, ...)
    THUtils.call(function()
        if self.missionInfo == nil then
            self.missionInfo = missionInfo
        end
    end)
    return superFunc(manager, xmlFile, missionInfo, baseDirectory, ...)
end
function THConversionSystem.hook_loadDefaultFruitTypes(superFunc, self, manager, ...)
    local function appendFunc(...)
        if self.missionInfo ~= nil then
            THUtils.call(function()
                local extXMLFilename = Utils.getFilename("map/fruitTypes.xml", self.modPath)
                local extXMLFile = THUtils.loadXMLFile("THFruitTypesXML", extXMLFilename)
                if extXMLFile ~= nil then
                    manager:loadFruitTypes(extXMLFile, self.missionInfo, false)
                    delete(extXMLFile)
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(manager, ...))
end