-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thConversionSystem.modName
local thEffectManager = g_thConversionSystem.effectManager
local thSettingsManager = g_thConversionSystem.thSettingsManager
local thFruitTypeManager = g_thConversionSystem.thFruitTypeManager
local thFillTypeManager = g_thConversionSystem.thFillTypeManager
THConversionCombine = {}
local debugFlagId = THDebugUtil.createDebugFlag("THConversionCombine")
THConversionCombine.debugFlagId = debugFlagId
THConversionCombine.MOD_NAME = thModName
THConversionCombine.SPEC_NAME = "conversionCombine"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THConversionCombine.SPEC_NAME, thModName, verbose)
end
THConversionCombine.getSpecTable = getSpecTable
function THConversionCombine.initSpecialization()
end
function THConversionCombine.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THCombine, specializations)
end
function THConversionCombine.registerEvents(vehicleType)
end
function THConversionCombine.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THConversionCombine)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", THConversionCombine)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", THConversionCombine)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", THConversionCombine)
end
function THConversionCombine.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "addCutterArea", THConversionCombine.addCutterArea)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setStrawPSEnabled", THConversionCombine.setStrawPSEnabled)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setChopperPSEnabled", THConversionCombine.setChopperPSEnabled)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "addCutterArea", THConversionCombine.addCutterArea)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processCombineChopperArea", THConversionCombine.processCombineChopperArea)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processCombineSwathArea", THConversionCombine.processCombineSwathArea)
end
function THConversionCombine:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.lastCutterParameters = {}
    specTable.strawParameters = {} -- needed for mp
    specTable.strawDirtyFlag = self:getNextDirtyFlag()
    specTable.functionParameters = {
        vehicle = self,
        isSwathProcessing = false,
        swapInputFruitType = false,
    }
    if not specTable.thIsInitialized then
        THUtils.setFunctionHook("FruitTypeManager", "getWindrowFillTypeIndexByFruitTypeIndex", nil, THConversionCombine.inj_getWindrowIndexByFruitTypeIndex, false, specTable.functionParameters)
        THUtils.setFunctionHook("FruitTypeManager", "getFruitTypeIndexByFillTypeIndex", nil, THConversionCombine.inj_getFruitTypeIndexByFillTypeIndex, false, specTable.functionParameters)
        specTable.thIsInitialized = true
    end
end
function THConversionCombine:onReadUpdateStream(streamId, timeStamp, connection)
    if connection:getIsServer() then
        local specTable = getSpecTable(self)
        if streamReadBool(streamId) then -- strawDirtyFlag
            local strawParams = THUtils.clearTable(specTable.strawParameters)
            strawParams.inputFruitType = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
            strawParams.strawFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
            strawParams.strawFactor = streamReadFloat32(streamId)
            strawParams.chopperFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
            strawParams.chopperFactor = streamReadFloat32(streamId)
            strawParams.useChoppedStraw = streamReadBool(streamId)
        end
    end
end
function THConversionCombine:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local specTable = getSpecTable(self)
        if streamWriteBool(streamId, bitAND(dirtyMask, specTable.strawDirtyFlag) ~= 0) then
            local strawParams = specTable.strawParameters
            local inputFruitType = strawParams.inputFruitType or FruitType.UNKNOWN
            local strawFillType = strawParams.strawFillType or FillType.UNKNOWN
            local strawFactor = strawParams.strawFactor or 1
            local chopperFillType = strawParams.chopperFillType or FillType.UNKNOWN
            local chopperFactor = strawParams.chopperFactor or 1
            local useChoppedStraw = strawParams.useChoppedStraw or false
            streamWriteUIntN(streamId, inputFruitType, FruitTypeManager.SEND_NUM_BITS)
            streamWriteUIntN(streamId, strawFillType, FillTypeManager.SEND_NUM_BITS)
            streamWriteFloat32(streamId, strawFactor)
            streamWriteUIntN(streamId, chopperFillType, FillTypeManager.SEND_NUM_BITS)
            streamWriteFloat32(streamId, chopperFactor)
            streamWriteBool(streamId, useChoppedStraw)
        end
    end
end
function THConversionCombine:onDelete()
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        specTable.functionParameters.vehicle = nil -- disables code injections
    end
end
function THConversionCombine:addCutterArea(superFunc, realArea, changedArea, inputFruitType, outputFillType, strawRatio, chopperType, farmId, cutterLoad, ...)
    local specTable = getSpecTable(self)
    local thCombineSpec = THCombine.getSpecTable(self)
    local vehicleName = self:getName()
    local cutterParams = specTable.lastCutterParameters
    local strawParams = specTable.strawParameters
    local isCutterAreaValid = false
    specTable.currentInputMoisture = nil
    THUtils.call(function()
        THUtils.clearTable(cutterParams)
        if inputFruitType ~= nil and inputFruitType ~= FruitType.UNKNOWN then
            local cutterVehicle = self:thAA_getCombineLastValidCutter()
            local thConversionCutterSpec = THConversionCutter.getSpecTable(cutterVehicle, false)
            if thConversionCutterSpec ~= nil then
                local modifierData = cutterVehicle:thAA_getCutterFruitTypeModifier(inputFruitType)
                if modifierData ~= nil then
                    cutterParams.strawFillType = modifierData.values.toStrawFillType
                    cutterParams.strawFactor = modifierData.values.toStrawFactor
                    cutterParams.chopperFillType = modifierData.values.toChopperFillType
                    cutterParams.chopperFactor = modifierData.values.toChopperFactor
                    cutterParams.useCutterRatio = modifierData.values.useCutterRatio
                    cutterParams.useChoppedStraw = modifierData.values.useChoppedStraw
                    cutterParams.fillTypeMoisture = modifierData.values.toMoisture
                end
            end
            if changedArea ~= nil and changedArea > 0
                and outputFillType ~= nil and outputFillType ~= FillType.UNKNOWN
            then
                isCutterAreaValid = true
                local thMoistureSystem = g_thAdvancedAddon:getModule("moistureSystem")
                local fillTypeMoisture = cutterParams.fillTypeMoisture or -1
                if thMoistureSystem ~= nil and fillTypeMoisture >= 0 then
                    specTable.currentInputMoisture = fillTypeMoisture
                end
            end
        end
    end)
    local function appendFunc(rAppliedDelta, ...)
        specTable.currentInputMoisture = nil
        THUtils.call(function()
            local baseCutterParams = thCombineSpec.lastCutterParameters
            local combineFillLevel = self:thAA_getCombineFillLevel()
            cutterParams.lastStrawArea = baseCutterParams.lastStrawArea or changedArea
            THUtils.clearTable(strawParams)
            if isCutterAreaValid and combineFillLevel > 0
                and rAppliedDelta ~= nil and rAppliedDelta > 0
            then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local inputFruitTypeName = thFruitTypeManager:getFruitTypeName(inputFruitType)
                    local outputFillTypeName = thFillTypeManager:getFillTypeName(outputFillType)
                    THUtils.displayMsg("%s >> Last cutter parameters:", vehicleName)
                    THUtils.displayMsg("- Input fruitType: [%s] (%s)", inputFruitType, inputFruitTypeName)
                    THUtils.displayMsg("- Output fillTYpe: [%s] (%s)", outputFillType, outputFillTypeName)
                    THUtils.printTable(cutterParams)
                end
                if self.isServer then
                    local strawFillType = cutterParams.strawFillType or FillType.UNKNOWN
                    local strawFactor = cutterParams.strawFactor or 1
                    local chopperFillType = cutterParams.chopperFillType or strawFillType
                    local chopperFactor = cutterParams.chopperFactor or strawFactor
                    local useChoppedStraw = THUtils.getNoNil(cutterParams.useChoppedStraw, true)
                    if cutterParams.lastStrawArea ~= nil then
                        local strawAreaFactor = cutterParams.lastStrawArea / changedArea
                    end
                    if cutterParams.useCutterRatio == false
                        and strawRatio ~= nil and strawRatio > 0
                    then
                        strawFactor = strawFactor / strawRatio
                        chopperFactor = chopperFactor / strawRatio
                    end
                    strawParams.inputFruitType = inputFruitType
                    strawParams.strawFillType = strawFillType
                    strawParams.strawFactor = strawFactor
                    strawParams.chopperFillType = chopperFillType
                    strawParams.chopperFactor = chopperFactor
                    strawParams.useChoppedStraw = useChoppedStraw
                    if THDebugUtil.getIsEnabled(debugFlagId) then
                        THUtils.displayMsg("- Straw parameters:")
                        THUtils.printTable(strawParams)
                    end
                    self:raiseDirtyFlags(specTable.strawDirtyFlag)
                end
            end
        end)
        return rAppliedDelta, ...
    end
    return appendFunc(superFunc(self, realArea, changedArea, inputFruitType, outputFillType, strawRatio, chopperType, farmId, cutterLoad, ...))
end
function THConversionCombine:setChopperPSEnabled(superFunc, isEnabled, ...)
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    local lastChopperPSEnabled = combineSpec.chopperPSenabled
    local function appendFunc(...)
        if self.isClient
            and isEnabled == true and isEnabled ~= lastChopperPSEnabled -- only when value has changed
            and combineSpec.chopperPSenabled == true -- only when newly enabled
        then
            THUtils.call(function()
                local strawParams = specTable.strawParameters
                local chopperEffects = combineSpec.chopperEffects
                local inputFruitTypeData = thFruitTypeManager:getFruitTypeData(strawParams.inputFruitType)
                if chopperEffects ~= nil
                    and inputFruitTypeData ~= nil and inputFruitTypeData.index ~= FruitType.UNKNOWN
                    and inputFruitTypeData.index == combineSpec.lastValidInputFruitType
                then
                    if strawParams.chopperFillType ~= nil and strawParams.chopperFillType ~= FillType.UNKNOWN then
                        local effectsFillType = inputFruitTypeData.fillTypeIndex
                        thEffectManager:stopEffects(chopperEffects)
                        thEffectManager:setFillType(chopperEffects, effectsFillType)
                        thEffectManager:startEffects(chopperEffects)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, isEnabled, ...))
end
function THConversionCombine:setStrawPSEnabled(superFunc, isEnabled, ...)
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    local lastStrawPSEnabled = combineSpec.strawPSenabled
    local function appendFunc(...)
        if self.isClient
            and isEnabled == true and lastStrawPSEnabled == false -- only when value has changed
            and combineSpec.strawPSenabled == true -- only when newly enabled
        then
            THUtils.call(function()
                local strawParams = specTable.strawParameters
                local strawEffects = combineSpec.strawEffects
                local inputFruitTypeData = thFruitTypeManager:getFruitTypeData(strawParams.inputFruitType)
                if strawEffects ~= nil
                    and inputFruitTypeData ~= nil and inputFruitTypeData.index ~= FruitType.UNKNOWN
                    and inputFruitTypeData.index == combineSpec.lastValidInputFruitType
                then
                    if strawParams.strawFillType ~= nil and strawParams.strawFillType ~= FillType.UNKNOWN then
                        local effectsFillType = inputFruitTypeData.fillTypeIndex
                        thEffectManager:stopEffects(strawEffects)
                        thEffectManager:setFillType(strawEffects, effectsFillType)
                        thEffectManager:startEffects(strawEffects)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, isEnabled, ...))
end
function THConversionCombine:processCombineChopperArea(superFunc, workArea, ...)
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    local workParams = combineSpec.workAreaParameters
    local function appendFunc(...)
        if not combineSpec.isSwathActive and workParams ~= nil then
            THUtils.call(function()
                local strawParams = specTable.strawParameters
                local litersToDrop = workParams.litersToDrop or 0
                local inputFruitType = strawParams.inputFruitType
                local isChoppedStrawEnabled = thSettingsManager:getSettingValue("useChoppedStraw") or false
                local useChoppedStraw = strawParams.useChoppedStraw
                if not isChoppedStrawEnabled then
                    useChoppedStraw = false
                end
                if useChoppedStraw == true and litersToDrop > 0
                    and inputFruitType ~= nil and inputFruitType ~= FruitType.UNKNOWN
                    and inputFruitType == combineSpec.lastValidInputFruitType
                then
                    local chopperFillType = strawParams.chopperFillType or strawParams.strawFillType or FillType.UNKNOWN
                    local chopperFactor = strawParams.chopperFactor or strawParams.strawFactor or 1
                    if chopperFillType == FillType.UNKNOWN then
                        local chopperFruitTypeInfo = thFruitTypeManager:getFruitTypeByFillType(workParams.dropFillType)
                        if chopperFruitTypeInfo ~= nil and chopperFruitTypeInfo ~= FillType.UNKNOWN
                            and chopperFruitTypeInfo.hasWindrow
                        then
                            local chopperWindrowInfo = thFruitTypeManager:getFruitTypeWindrowFillType(chopperFruitTypeInfo.index)
                            if chopperWindrowInfo ~= nil and chopperWindrowInfo.index ~= FillType.UNKNOWN then
                                chopperFillType = chopperWindrowInfo.index
                            end
                        end
                    end
                    if chopperFactor > 0
                        and chopperFillType ~= nil and chopperFillType ~= FillType.UNKNOWN
                    then
                        if THDebugUtil.getIsEnabled(debugFlagId) then
                            local vehicleName = self:getName()
                            local inputFruitTypeName = thFruitTypeManager:getFruitTypeName(inputFruitType)
                            local chopperFillTypeName = thFillTypeManager:getFillTypeName(chopperFillType)
                            THUtils.displayMsg("%s >> Combine chopper parameters", vehicleName)
                            THUtils.displayMsg("- Input fruitType: [%s] (%s)", inputFruitType, inputFruitTypeName)
                            THUtils.displayMsg("- Chopper fillType: [%s] (%s)", chopperFillType, chopperFillTypeName)
                            THUtils.displayMsg("- Chopper factor: %s", chopperFactor)
                            THUtils.printTable(strawParams)
                        end
                        local droppedLiters = THDensityMapHeightConversionUtil.addChoppedStrawArea(self, chopperFillType, litersToDrop * chopperFactor, workArea, nil, true)
                        if droppedLiters ~= nil then
                            workParams.droppedLiters = droppedLiters / chopperFactor
                            workParams.isChopperEffectEnabled = 500
                        end
                        self:raiseActive()
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, workArea, ...))
end
function THConversionCombine:processCombineSwathArea(superFunc, workArea, ...)
    local specTable = getSpecTable(self)
    local combineSpec = self.spec_combine
    local workParams = combineSpec.workAreaParameters
    local oldLitersToDrop = nil
    local strawFactor = 1
    local prependSuccess = false
    if combineSpec.isSwathActive and workParams ~= nil
        and workParams.litersToDrop ~= nil and workParams.litersToDrop > 0
    then
        THUtils.call(function()
            local strawParams = specTable.strawParameters
            local fruitTypeData = thFruitTypeManager:getFruitTypeData(strawParams.inputFruitType)
            if fruitTypeData ~= nil and fruitTypeData.fillTypeIndex == workParams.dropFillType then
                specTable.functionParameters.isSwathProcessing = true
                strawFactor = strawParams.strawFactor or 1
                oldLitersToDrop = workParams.litersToDrop
                workParams.litersToDrop = oldLitersToDrop * strawFactor
                prependSuccess = true
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local vehicleName = self:getName()
                    local strawFillTypeName = thFillTypeManager:getFillTypeName(strawParams.strawFillType)
                    THUtils.displayMsg("%s >> Combine straw parameters", vehicleName)
                    THUtils.displayMsg("- Input fruitType: [%s] (%s)", fruitTypeData.index, fruitTypeData.name)
                    THUtils.displayMsg("- Straw fillType: [%s] (%s)", strawParams.strawFillType, strawFillTypeName)
                    THUtils.displayMsg("- Straw factor: %s", strawFactor)
                    THUtils.printTable(strawParams)
                end
            end
        end)
    end
    local function appendFunc(...)
        if specTable.functionParameters.isSwathProcessing then
            specTable.functionParameters.isSwathProcessing = false
            if prependSuccess then -- implies other checks
                THUtils.call(function()
                    if workParams ~= nil and workParams.litersToDrop ~= nil then
                        if workParams.droppedLiters ~= nil then
                            if strawFactor <= 0 then
                                workParams.droppedLiters = 0
                            else
                                workParams.droppedLiters = workParams.droppedLiters / strawFactor
                            end
                        end
                        workParams.litersToDrop = oldLitersToDrop -- restore
                    end
                end)
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, workArea, ...))
end
function THConversionCombine:hook_thAA_getCombineInputMoisture(superFunc, ...)
    local specTable = getSpecTable(self, false)
    local function appendFunc(rMoisture, ...)
        if specTable ~= nil and specTable.currentInputMoisture ~= nil then
            rMoisture = specTable.currentInputMoisture
        end
        return rMoisture, ...
    end
    return appendFunc(superFunc(self, ...))
end
function THConversionCombine:hook_onStartWorkAreaProcessing(superFunc, dt, ...)
    local specTable = getSpecTable(self, false)
    local combineSpec = self.spec_combine
    local workParams = combineSpec.workAreaParameters
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                local strawParams = specTable.strawParameters
                local combineFillLevel = self:thAA_getCombineFillLevel()
                local inputFruitTypeData = thFruitTypeManager:getFruitTypeData(strawParams.inputFruitType)
                if combineFillLevel > 0
                    and inputFruitTypeData ~= nil and inputFruitTypeData.index ~= FruitType.UNKNOWN
                    and inputFruitTypeData.index == combineSpec.lastValidInputFruitType
                then
                    workParams.dropFillType = inputFruitTypeData.fillTypeIndex
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THConversionCombine:hook_onUpdate(superFunc, ...)
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        specTable.functionParameters.swapInputFruitType = true
    end
    local function appendFunc(...)
        if specTable ~= nil then
            specTable.functionParameters.swapInputFruitType = false
        end
        return ...
    end
    return appendFunc(superFunc(self, ...))
end
function THConversionCombine:hook_actionEventToggleChopper(superFunc, ...)
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        specTable.functionParameters.swapInputFruitType = true
    end
    local function appendFunc(...)
        if specTable ~= nil then
            specTable.functionParameters.swapInputFruitType = false
        end
        return ...
    end
    return appendFunc(superFunc(self, ...))
end
function THConversionCombine.inj_getWindrowIndexByFruitTypeIndex(superFunc, extraArgs, manager, fruitTypeIndex, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and extraArgs.isSwathProcessing
            and fruitTypeIndex ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN
        then
            local strawFillType = FillType.UNKNOWN
            local allowUpdate = false
            THUtils.call(function()
                local combineSpec = self.spec_combine
                local workParams = combineSpec.workAreaParameters
                local strawParams = specTable.strawParameters
                strawFillType = strawParams.strawFillType or FillType.UNKNOWN
                if strawFillType ~= FillType.UNKNOWN
                    and workParams ~= nil and workParams.dropFillType ~= nil
                    and workParams.dropFillType ~= FillType.UNKNOWN
                then
                    local dropFruitTypeInfo = manager:getFruitTypeByFillTypeIndex(workParams.dropFillType)
                    if dropFruitTypeInfo ~= nil and dropFruitTypeInfo.index == fruitTypeIndex then
                        allowUpdate = true
                    end
                end
            end)
            if allowUpdate then
                local function appendFunc(rWindrowIndex, ...)
                    rWindrowIndex = strawFillType
                    return rWindrowIndex, ...
                end
                return appendFunc(superFunc(manager, fruitTypeIndex, ...))
            end
        end
    end
    return superFunc(manager, fruitTypeIndex, ...)
end
function THConversionCombine.inj_getFruitTypeIndexByFillTypeIndex(superFunc, extraArgs, manager, fillTypeIndex, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and extraArgs.swapInputFruitType then
            local fruitTypeIndex = FruitType.UNKNOWN
            THUtils.call(function()
                local combineSpec = self.spec_combine
                local combineFillLevel, combineFillType = self:thAA_getCombineFillLevel()
                if combineFillLevel <= 0 or combineFillType == fillTypeIndex then
                    if combineSpec.lastValidInputFruitType ~= nil
                        and combineSpec.lastValidInputFruitType ~= FruitType.UNKNOWN
                    then
                        fruitTypeIndex = combineSpec.lastValidInputFruitType
                    else
                        fruitTypeIndex = specTable.strawParameters.inputFruitType or FruitType.UNKNOWN
                    end
                end
            end)
            if fruitTypeIndex ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN then
                local function appendFunc(rFruitTypeIndex, ...)
                    rFruitTypeIndex = fruitTypeIndex
                    return rFruitTypeIndex, ...
                end
                return appendFunc(superFunc(manager, fillTypeIndex, ...))
            end
        end
    end
    return superFunc(manager, fillTypeIndex, ...)
end
g_thConversionSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Combine", "onStartWorkAreaProcessing", THConversionCombine)
    THUtils.setFunctionHook("Combine", "onUpdate", THConversionCombine)
    THUtils.setFunctionHook("Combine", "actionEventToggleChopper", THConversionCombine)
    if THMoistureCombine ~= nil then
        THUtils.setFunctionHook(THMoistureCombine, "thAA_setCombineInputMoisture", THConversionCombine)
    end
end)