-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thConversionSystem.modName
local thFruitTypeManager = g_thConversionSystem.thFruitTypeManager
local thFillTypeManager = g_thConversionSystem.thFillTypeManager
THConversionCutter = {}
local debugFlagId = THDebugUtil.createDebugFlag("THConversionCutter")
THConversionCutter.debugFlagId = debugFlagId
THConversionCutter.MOD_NAME = thModName
THConversionCutter.SPEC_NAME = "conversionCutter"
THConversionCutter.CONVERTER = {
    UNKNOWN = "UNKNOWN",
    NONE = "NONE",
    DEFAULT = THConversionSystem.CONVERTER.DEFAULT
}
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THConversionCutter.SPEC_NAME, thModName, verbose)
end
THConversionCutter.getSpecTable = getSpecTable
function THConversionCutter.initSpecialization()
end
function THConversionCutter.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THCutter, specializations)
end
function THConversionCutter.registerEvents(vehicleType)
end
function THConversionCutter.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THConversionCutter)
    SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", THConversionCutter)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", THConversionCutter)
end
function THConversionCutter.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getCutterConverter", THConversionCutter.thAA_getCutterConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_setCutterConverter", THConversionCutter.thAA_setCutterConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_addCutterConverter", THConversionCutter.thAA_addCutterConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_updateCutterConverter", THConversionCutter.thAA_updateCutterConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getCutterFruitTypeModifier", THConversionCutter.thAA_getCutterFruitTypeModifier)
end
function THConversionCutter.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processCutterArea", THConversionCutter.processCutterArea)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processPickupCutterArea", THConversionCutter.processPickupCutterArea)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "readCutterFromStream", THConversionCutter.readCutterFromStream)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "writeCutterToStream", THConversionCutter.writeCutterToStream)
end
local function getCutterAllowedFillTypes(self)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    local combineVehicle = workParams.combineVehicle
    if combineVehicle ~= nil and combineVehicle.thAA_getCombineFillLevel ~= nil then
        local combineFillLevel, combineFillType, combineFillUnit = combineVehicle:thAA_getCombineFillLevel()
        local supportedFillTypes = combineVehicle:thAA_getCombineSupportedFillTypes(true)
        local requiredFillType = nil
        if combineFillLevel ~= nil and combineFillLevel > 0
            and combineFillType ~= nil and combineFillType ~= FillType.UNKNOWN
        then
            local combineFillLevelThreshold = combineVehicle:getFillTypeChangeThreshold(combineFillUnit)
            if combineFillLevel >= combineFillLevelThreshold then
                requiredFillType = combineFillType
            end
        end
        return supportedFillTypes, requiredFillType
    end
end
function THConversionCutter:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.fruitTypeConverters = {
        byId = {},
        byIndex = {},
        defaultName = THConversionCutter.CONVERTER.UNKNOWN,
        defaultIndex = 0,
        currentIndex = 0,
        lastIndex = 0
    }
    specTable.defaultValues = {
        fruitTypes = {},
        fillTypes = {}
    }
    specTable.fruitTypesArray = {}
    specTable.fruitTypeMapping = {}
    specTable.activeFruitTypes = {}
    specTable.activeFruitTypesArray = {}
    specTable.cachedFruitTypes = {}
    specTable.fillTypesArray = {}
    specTable.fillTypeMapping = {}
    specTable.harvestParameters = {}
    specTable.lastInputFillType = FillType.UNKNOWN
    specTable.conversionFillTypes = {}
    specTable.lastPickupFruitType = FruitType.UNKNOWN
    specTable.lastPickupFillType = FillType.UNKNOWN
    specTable.hasPickupFunction = false
    specTable.functionParameters = {
        vehicle = self,
        isCutterProcessing = false,
        isPickupProcessing = false
    }
    if not specTable.thIsInitialized then
        THUtils.setFunctionHook("FSDensityMapUtil", "cutFruitArea", nil, THConversionCutter.inj_cutFruitArea, false, specTable.functionParameters)
        THUtils.setFunctionHook("FSDensityMapUtil", "setGroundTypeLayerArea", nil, THConversionCutter.inj_setGroundTypeLayerArea, false, specTable.functionParameters)
        THUtils.setFunctionHook("FruitTypeManager", "getWindrowFillTypeIndexByFruitTypeIndex", nil, THConversionCutter.inj_getWindrowFillTypeIndexByFruitTypeIndex, false, specTable.functionParameters)
        specTable.thIsInitialized = true
    end
end
function THConversionCutter:onPostLoad(savegame)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local vehicleName = self:getName()
    local workAreaSpec = self.spec_workArea
    local function addValidFruitType(pFruitTypeIndex)
        if specTable.fruitTypeMapping[pFruitTypeIndex] == nil then
            table.insert(specTable.fruitTypesArray, pFruitTypeIndex)
            specTable.fruitTypeMapping[pFruitTypeIndex] = #specTable.fruitTypesArray
        end
    end
    local function addValidFillType(pFillTypeIndex)
        if specTable.fillTypeMapping[pFillTypeIndex] == nil then
            table.insert(specTable.fillTypesArray, pFillTypeIndex)
            specTable.fillTypeMapping[pFillTypeIndex] = #specTable.fillTypesArray
        end
    end
    if cutterSpec.fruitTypes ~= nil then
        specTable.defaultValues.fruitTypes = cutterSpec.fruitTypes
        for _, fruitTypeIndex in ipairs(cutterSpec.fruitTypes) do
            addValidFruitType(fruitTypeIndex)
        end
    end
    if cutterSpec.fillTypes ~= nil then
        specTable.defaultValues.fillTypes = cutterSpec.fillTypes
        for _, fillTypeIndex in ipairs(cutterSpec.fillTypes) do
            addValidFillType(fillTypeIndex)
        end
    end
    local converterName = THUtils.getXMLValue(self.xmlFile, "vehicle.cutter", "#fruitTypeConverter")
    local defaultConverter = self:thAA_addCutterConverter(THConversionCutter.CONVERTER.NONE)
    local isPickup = false
    if converterName ~= nil then
        local realConverter = self:thAA_addCutterConverter(converterName, true)
        if realConverter ~= nil then
            defaultConverter = realConverter
        end
    end
    if defaultConverter == nil then
        THUtils.stackMsg("Could not create default cutter converter")
        converterName = THConversionCutter.CONVERTER.UNKNOWN
    else
        converterName = defaultConverter.name
        specTable.fruitTypeConverters.defaultName = converterName
        specTable.fruitTypeConverters.defaultIndex = defaultConverter.index
        self:thAA_setCutterConverter(defaultConverter.index, true)
    end
    if workAreaSpec ~= nil then
        local workAreasArray = workAreaSpec.workAreas
        if workAreasArray ~= nil then
            for _, workArea in ipairs(workAreasArray) do
                if workArea.functionName == "processPickupCutterArea" then
                    isPickup = true
                    break
                end
            end
        end
    end
    specTable.hasPickupFunction = isPickup == true
    for _, fruitTypeIndex in ipairs(specTable.fruitTypesArray) do
        local fruitTypeConversionData = g_thConversionSystem:getFruitTypeConversionData(fruitTypeIndex)
        if fruitTypeConversionData ~= nil then
            if isPickup then
                for _, pickupData in ipairs(fruitTypeConversionData.pickupFillTypes) do
                    addValidFillType(pickupData.fillTypeIndex)
                    g_thConversionSystem:getFruitTypeConversionFillTypes(fruitTypeIndex, nil, converterName, pickupData.fillTypeIndex, isPickup, specTable.conversionFillTypes)
                end
            end
        end
        for _, otherConverter in ipairs(specTable.fruitTypeConverters.byIndex) do
            local converterInfo = otherConverter.data[fruitTypeIndex]
            if converterInfo ~= nil and converterInfo.fillTypeIndex ~= nil then
                g_thConversionSystem:getFruitTypeConversionFillTypes(fruitTypeIndex, nil, otherConverter.name, converterInfo.fillTypeIndex, isPickup, specTable.conversionFillTypes)
            end
        end
    end
    for fillTypeIndex in pairs(specTable.conversionFillTypes) do
        addValidFillType(fillTypeIndex)
    end
    cutterSpec.fruitTypes = specTable.fruitTypesArray
    cutterSpec.fillTypes = specTable.fillTypesArray
    if THDebugUtil.getIsEnabled(debugFlagId) then
        THUtils.displayMsg("%s >> Cutter information:", vehicleName)
        THUtils.displayMsg("- Converter name: %s", converterName)
        THUtils.displayMsg("- Has pickup function: %s", specTable.hasPickupFunction)
        THUtils.displayMsg("")
        THUtils.displayMsg("- Valid fruitTypes:")
        for _, fruitTypeIndex in ipairs(cutterSpec.fruitTypes) do
            local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
            THUtils.displayMsg("[%s]: %s", fruitTypeIndex, fruitTypeName)
        end
        THUtils.displayMsg("")
        THUtils.displayMsg("- Valid fillTypes:")
        for _, fillTypeIndex in ipairs(cutterSpec.fillTypes) do
            local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
            THUtils.displayMsg("[%s]: %s", fillTypeIndex, fillTypeName)
        end
        THUtils.displayMsg("")
        if specTable.conversionFillTypes ~= nil
            and next(specTable.conversionFillTypes) ~= nil
        then
            THUtils.displayMsg("- Conversion fillTypes:")
            for fillTypeIndex in pairs(specTable.conversionFillTypes) do
                local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
                THUtils.displayMsg("[%s]: %s", fillTypeIndex, fillTypeName)
            end
            THUtils.displayMsg("")
        end
        if cutterSpec.fruitTypeConverters ~= nil
            and next(cutterSpec.fruitTypeConverters) ~= nil
        then
            THUtils.displayMsg("- Default converters:")
            for fruitTypeIndex, converterInfo in pairs(cutterSpec.fruitTypeConverters) do
                local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                THUtils.displayMsg("[%s]: %s", fruitTypeIndex, fruitTypeName)
                THUtils.printTable(converterInfo)
            end
        end
        if #specTable.fruitTypeConverters.byIndex > 0 then
            THUtils.displayMsg("- Extended converters:")
            for _, otherConverter in ipairs(specTable.fruitTypeConverters.byIndex) do
                THUtils.printTable(otherConverter)
                for fruitTypeIndex, otherConverterInfo in pairs(otherConverter.data) do
                    local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                    THUtils.displayMsg("[%s]: %s", fruitTypeName, fruitTypeIndex)
                    THUtils.printTable(otherConverterInfo)
                end
            end
        end
    end
end
function THConversionCutter:onDelete()
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        specTable.functionParameters.vehicle = nil -- disables code injections
    end
end
function THConversionCutter:thAA_getCutterConverter(converterId)
    local specTable = getSpecTable(self)
    if converterId == nil then
        return specTable.fruitTypeConverters.byIndex[specTable.fruitTypeConverters.currentIndex]
    else
        local idType = type(converterId)
        if idType == "number" then
            return specTable.fruitTypeConverters.byIndex[converterId]
        elseif idType == "string" then
            return specTable.fruitTypeConverters.byId[converterId:upper()]
        end
    end
end
function THConversionCutter:thAA_setCutterConverter(converterId, noEventSend)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local vehicleName = self:getName()
    if converterId == nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "converterId", converterId)
        return nil, false
    end
    local newConverter = self:thAA_getCutterConverter(converterId)
    local isConverterDirty = false
    if newConverter == nil then
        THUtils.warningMsg("%s >> Could not find converter (%s)", vehicleName, converterId)
        return nil, false
    end
    if specTable.fruitTypeConverters.currentIndex ~= newConverter.index
        or specTable.fruitTypeConverters.lastIndex ~= newConverter.index
        or cutterSpec.fruitTypeConverters ~= newConverter.data
    then
        if THDebugUtil.getIsEnabled(debugFlagId) then
            local oldConverter = self:thAA_getCutterConverter(specTable.fruitTypeConverters.lastIndex)
            if oldConverter == nil then
                THUtils.displayMsg("%s >> Cutter converter switched to %q", vehicleName, newConverter.name)
            elseif oldConverter.index ~= newConverter.index then
                THUtils.displayMsg("%s >> Cutter converter switched from %q to %q", vehicleName, oldConverter.name, newConverter.name)
            end
        end
        specTable.fruitTypeConverters.currentIndex = newConverter.index
        specTable.fruitTypeConverters.lastIndex = newConverter.index
        cutterSpec.fruitTypeConverters = newConverter.data
        if self.isServer and noEventSend ~= true then
            self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
        end
        isConverterDirty = true
    end
    return newConverter, isConverterDirty
end
function THConversionCutter:thAA_addCutterConverter(name, useCurrent)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local vehicleName = self:getName()
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name) then
        local converterId = name:upper()
        local newConverter = self:thAA_getCutterConverter(converterId)
        if newConverter == nil then
            newConverter = {
                name = converterId,
                index = #specTable.fruitTypeConverters.byIndex + 1,
                data = {},
                lastData = {},
                realData = {},
                pickupFillTypes = {},
                modifiers = {}
            }
            local defaultConverterData = thFruitTypeManager:getFruitTypeConverterData(THConversionCutter.CONVERTER.DEFAULT)
            if defaultConverterData ~= nil then
                for _, fruitTypeIndex in ipairs(specTable.fruitTypesArray) do
                    if defaultConverterData[fruitTypeIndex] ~= nil then
                        if newConverter.realData[fruitTypeIndex] == nil then
                            newConverter.realData[fruitTypeIndex] = {}
                        end
                        if newConverter.data[fruitTypeIndex] == nil then
                            newConverter.data[fruitTypeIndex] = {}
                        end
                        for key, value in pairs(defaultConverterData[fruitTypeIndex]) do
                            newConverter.realData[fruitTypeIndex][key] = value
                            newConverter.data[fruitTypeIndex][key] = value
                        end
                    end
                end
            end
            if converterId ~= THConversionCutter.CONVERTER.NONE then
                local realConverterData = thFruitTypeManager:getFruitTypeConverterData(converterId)
                if useCurrent == true and cutterSpec.fruitTypeConverters ~= nil then
                    realConverterData = cutterSpec.fruitTypeConverters
                end
                if realConverterData ~= nil then
                    for fruitTypeIndex, realConverterInfo in pairs(realConverterData) do
                        if newConverter.realData[fruitTypeIndex] == nil then
                            newConverter.realData[fruitTypeIndex] = {}
                        end
                        if newConverter.data[fruitTypeIndex] == nil then
                            newConverter.data[fruitTypeIndex] = {}
                        end
                        for key, value in pairs(realConverterInfo) do
                            newConverter.realData[fruitTypeIndex][key] = value
                            newConverter.data[fruitTypeIndex][key] = value
                        end
                    end
                end
            end
            specTable.fruitTypeConverters.byIndex[newConverter.index] = newConverter
            specTable.fruitTypeConverters.byId[newConverter.name] = newConverter
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THUtils.displayMsg("%s >> Added cutter converter %q", vehicleName, newConverter.name)
            end
        end
        return newConverter
    end
end
function THConversionCutter:thAA_updateCutterConverter(inputFruitType, pickupFillType, noEventSend)
    local cutterSpec = self.spec_cutter
    local vehicleName = self:getName()
    local currentConverter = self:thAA_getCutterConverter()
    if currentConverter ~= nil then
        local converterInfo = currentConverter.data[inputFruitType]
        local realConverterInfo = currentConverter.realData[inputFruitType]
        local isConverterDirty = false
        if realConverterInfo ~= nil then
            if converterInfo == nil then
                converterInfo = {}
                currentConverter.data[inputFruitType] = converterInfo
            end
            for key, value in pairs(realConverterInfo) do
                converterInfo[key] = value
            end
            pickupFillType = pickupFillType or FillType.UNKNOWN
            if pickupFillType ~= FillType.UNKNOWN then
                converterInfo.fillTypeIndex = pickupFillType
                converterInfo.conversionFactor = 1
                converterInfo.windrowConversionFactor = 1
            end
            if pickupFillType ~= currentConverter.pickupFillTypes[inputFruitType] then
                currentConverter.pickupFillTypes[inputFruitType] = pickupFillType
                isConverterDirty = true
            end
            local modifierIndex = currentConverter.modifiers[inputFruitType]
            local modifierData = g_thConversionSystem:getFruitTypeModifierDataByIndex(inputFruitType, modifierIndex)
            if modifierData ~= nil then
                if modifierData.values.toFillType ~= nil then
                    converterInfo.fillTypeIndex = modifierData.values.toFillType
                    converterInfo.conversionFactor = 1
                    converterInfo.windrowConversionFactor = 1
                end
                if modifierData.values.toFillTypeFactor ~= nil then
                    converterInfo.conversionFactor = modifierData.values.toFillTypeFactor
                    converterInfo.windrowConversionFactor = modifierData.values.toFillTypeFactor
                end
            end
            if currentConverter.lastData[inputFruitType] == nil then
                currentConverter.lastData[inputFruitType] = {}
            end
            for key, value in pairs(converterInfo) do
                if currentConverter.lastData[inputFruitType][key] ~= value then
                    currentConverter.lastData[inputFruitType][key] = value
                    isConverterDirty = true
                end
            end
            if self.isServer and cutterSpec.currentInputFruitType == inputFruitType then
                if cutterSpec.currentOutputFillType ~= converterInfo.fillTypeIndex
                    or cutterSpec.currentConversionFactor ~= converterInfo.conversionFactor
                then
                    cutterSpec.currentInputFruitType = FruitType.UNKNOWN
                    isConverterDirty = true
                end
            end
            if isConverterDirty then
                if self.isServer and noEventSend ~= true then
                    self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
                end
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local inputFruitTypeName = thFruitTypeManager:getFruitTypeName(inputFruitType)
                    local inputFillTypeName = thFillTypeManager:getFillTypeName(realConverterInfo.fillTypeIndex)
                    THUtils.displayMsg("%s >> Cutter conversion information:", vehicleName)
                    THUtils.displayMsg("- Converter name: %s", currentConverter.name)
                    THUtils.displayMsg("- Input fruitType: [%s] (%s)", inputFruitType, inputFruitTypeName)
                    THUtils.displayMsg("- Input fillType: [%s] (%s)", realConverterInfo.fillTypeIndex, inputFillTypeName)
                    THUtils.printTable(converterInfo)
                    if modifierData ~= nil then
                        THUtils.displayMsg("- Modifier Values:")
                        THUtils.printTable(modifierData.values)
                        THUtils.displayMsg("- Modifier Conditions:")
                        THUtils.printTable(modifierData.conditions, 1)
                    end
                end
            end
            return true, isConverterDirty
        end
    end
    return false, false
end
function THConversionCutter:thAA_getCutterFruitTypeModifier(fruitType)
    local currentConverter = self:thAA_getCutterConverter()
    local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
    if currentConverter ~= nil and fruitTypeInfo ~= nil then
        local modifierIndex = currentConverter.modifiers[fruitTypeInfo.index]
        if modifierIndex ~= nil then
            return g_thConversionSystem:getFruitTypeModifierDataByIndex(fruitTypeInfo.index, modifierIndex)
        end
    end
end
function THConversionCutter:processCutterArea(superFunc, workArea, ...)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    specTable.functionParameters.isCutterProcessing = true
    local function appendFunc(...)
        specTable.functionParameters.isCutterProcessing = false
        THUtils.call(function()
            local fruitTypeIndex = workParams.lastFruitType
            if fruitTypeIndex ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN then
                local modifierData = self:thAA_getCutterFruitTypeModifier(fruitTypeIndex)
                if modifierData ~= nil then
                    if modifierData.values.toChopperType ~= nil then
                        workParams.lastChopperValue = modifierData.values.toChopperValue
                    end
                end
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, workArea, ...))
end
function THConversionCutter:processPickupCutterArea(superFunc, workArea, dt, ...)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    local oldFruitTypesToUse = nil
    local areFruitTypesChanged = false
    if workArea ~= nil and workParams ~= nil
        and workParams.combineVehicle ~= nil
        and workParams.fruitTypesToUse ~= nil
    then
        THUtils.call(function()
            oldFruitTypesToUse = workParams.fruitTypesToUse
            if self.isServer then
                local lastSentFruitType = specTable.lastPickupFruitType
                local lastSentFillType = specTable.lastPickupFillType
                specTable.lastPickupFruitType = FruitType.UNKNOWN
                specTable.lastPickupFillType = FillType.UNKNOWN
                local wsx, wsy, wsz = getWorldTranslation(workArea.start)
                local wwx, wwy, wwz = getWorldTranslation(workArea.width)
                local whx, why, whz = getWorldTranslation(workArea.height)
                local lsx, lsy, lsz, lex, ley, lez, lineRadius = DensityMapHeightUtil.getLineByAreaDimensions(wsx, wsy, wsz, wwx, wwy, wwz, whx, why, whz)
                if lineRadius ~= nil and lineRadius > 0 then
                    local lineFillType = DensityMapHeightUtil.getFillTypeAtLine(lsx, lsy, lsz, lex, ley, lez, lineRadius)
                    if lineFillType ~= nil and lineFillType ~= FillType.UNKNOWN then
                        for _, fruitTypeIndex in pairs(workParams.fruitTypesToUse) do
                            local pickupData = g_thConversionSystem:getFruitTypePickupData(fruitTypeIndex, lineFillType)
                            if pickupData ~= nil then
                                specTable.lastPickupFruitType = fruitTypeIndex
                                specTable.lastPickupFillType = lineFillType
                                if specTable.lastPickupFruitType ~= lastSentFruitType
                                    or specTable.lastPickupFillType ~= lastSentFillType
                                then
                                    self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
                                end
                                break
                            end
                        end
                    end
                end
            end
            if specTable.lastPickupFruitType ~= nil and specTable.lastPickupFruitType ~= FruitType.UNKNOWN then
                workParams.fruitTypesToUse = THUtils.clearTable(specTable.cachedFruitTypes)
                table.insert(workParams.fruitTypesToUse, specTable.lastPickupFruitType)
                areFruitTypesChanged = true
            end
        end)
    end
    specTable.functionParameters.isPickupProcessing = true
    local function appendFunc(...)
        specTable.functionParameters.isPickupProcessing = false
        if areFruitTypesChanged then
            workParams.fruitTypesToUse = oldFruitTypesToUse
            if self.isServer then
                cutterSpec.currentInputFillType = specTable.lastPickupFillType
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, workArea, dt, ...))
end
function THConversionCutter:readCutterFromStream(superFunc, streamId, connection, ...)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local currentConverter = nil
    if connection:getIsServer() then
        local converterIndex = streamReadUIntN(streamId, 4)
        THUtils.call(function()
            if converterIndex > 0 then
                currentConverter = self:thAA_setCutterConverter(converterIndex, true)
            end
        end)
        local numFruitTypes = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
        if numFruitTypes > 0 then
            for _ = 1, numFruitTypes do
                local fruitTypeIndex = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
                local modifierIndex = streamReadUIntN(streamId, THConversionSystem.MODIFIER_SEND_BITS)
                local pickupFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
                THUtils.call(function()
                    if currentConverter ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN then
                        currentConverter.modifiers[fruitTypeIndex] = modifierIndex
                        self:thAA_updateCutterConverter(fruitTypeIndex, pickupFillType, true)
                    end
                end)
            end
        end
    end
    local function appendFunc(...)
        if connection:getIsServer() then
            specTable.lastPickupFruitType = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
            specTable.lastPickupFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
            if streamReadBool(streamId) then
                cutterSpec.currentInputFillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
            end
        end
        THUtils.call(function()
            if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                local vehicleName = self:getName()
                local pickupFruitTypeName = thFruitTypeManager:getFruitTypeName(specTable.lastPickupFruitType)
                local pickupFillTypeName = thFillTypeManager:getFillTypeName(specTable.lastPickupFillType)
                THUtils.displayMsg("%s >> Current readStream values:", vehicleName)
                THUtils.displayMsg("- Last pickup fruitType: [%s] (%s)", pickupFruitTypeName, specTable.lastPickupFruitType)
                THUtils.displayMsg("- Last pickup fillType: [%s] (%s)", pickupFillTypeName, specTable.lastPickupFillType)
                if cutterSpec.currentInputFillType ~= nil then
                    local inputFillTypeName = thFillTypeManager:getFillTypeName(cutterSpec.currentInputFillType)
                    THUtils.displayMsg("- Current input fillType: [%s] (%s)", inputFillTypeName, cutterSpec.currentInputFillType)
                end
                if currentConverter ~= nil then
                    THUtils.displayMsg("- Current converter name: %s", currentConverter.name)
                    THUtils.displayMsg("- Current converter index: %s", currentConverter.index)
                    THUtils.displayMsg("- Current converter info:")
                    THUtils.printTable(currentConverter, 2)
                end
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THConversionCutter:writeCutterToStream(superFunc, streamId, connection, ...)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local currentConverter = nil
    local converterIndex = 0
    if not connection:getIsServer() then
        THUtils.call(function()
            currentConverter = self:thAA_getCutterConverter()
            if currentConverter ~= nil then
                converterIndex = currentConverter.index or 0
            end
        end)
        local numFruitTypes = #specTable.activeFruitTypesArray
        streamWriteUIntN(streamId, converterIndex, 4)
        streamWriteUIntN(streamId, numFruitTypes, FruitTypeManager.SEND_NUM_BITS)
        if numFruitTypes > 0 then
            for i = 1, numFruitTypes do
                local fruitTypeIndex = specTable.activeFruitTypesArray[i]
                local modifierIndex = 0
                local pickupFillType = FillType.UNKNOWN
                if currentConverter ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN then
                    modifierIndex = currentConverter.modifiers[fruitTypeIndex] or 0
                    pickupFillType = currentConverter.pickupFillTypes[fruitTypeIndex] or FillType.UNKNOWN
                end
                streamWriteUIntN(streamId, fruitTypeIndex, FruitTypeManager.SEND_NUM_BITS)
                streamWriteUIntN(streamId, modifierIndex, THConversionSystem.MODIFIER_SEND_BITS)
                streamWriteUIntN(streamId, pickupFillType, FillTypeManager.SEND_NUM_BITS)
            end
        end
    end
    local function appendFunc(...)
        if not connection:getIsServer() then
            streamWriteUIntN(streamId, specTable.lastPickupFruitType or FruitType.UNKNOWN, FruitTypeManager.SEND_NUM_BITS)
            streamWriteUIntN(streamId, specTable.lastPickupFillType or FillType.UNKNOWN, FillTypeManager.SEND_NUM_BITS)
            if streamWriteBool(streamId, cutterSpec.useWindrow == true) then
                streamWriteUIntN(streamId, specTable.lastInputFillType or FillType.UNKNOWN, FillTypeManager.SEND_NUM_BITS)
            end
        end
        THUtils.call(function()
            if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                local vehicleName = self:getName()
                local pickupFruitTypeName = thFruitTypeManager:getFruitTypeName(specTable.lastPickupFruitType)
                local pickupFillTypeName = thFillTypeManager:getFillTypeName(specTable.lastPickupFillType)
                THUtils.displayMsg("%s >> Current writeStream values:", vehicleName)
                THUtils.displayMsg("- Last pickup fruitType: [%s] (%s)", pickupFruitTypeName, specTable.lastPickupFruitType)
                THUtils.displayMsg("- Last pickup fillType: [%s] (%s)", pickupFillTypeName, specTable.lastPickupFillType)
                if cutterSpec.currentInputFillType ~= nil then
                    local inputFillTypeName = thFillTypeManager:getFillTypeName(cutterSpec.currentInputFillType)
                    THUtils.displayMsg("- Current input fillType: [%s] (%s)", inputFillTypeName, cutterSpec.currentInputFillType)
                end
                if currentConverter ~= nil then
                    THUtils.displayMsg("- Current converter name: %s", currentConverter.name)
                    THUtils.displayMsg("- Current converter index: %s", currentConverter.index)
                    THUtils.displayMsg("- Current converter info:")
                    THUtils.printTable(currentConverter, 2)
                end
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THConversionCutter:hook_onStartWorkAreaProcessing(superFunc, dt, ...)
    local specTable = getSpecTable(self, false)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    local currentConverter = nil
    if specTable ~= nil then
        THUtils.call(function()
            THUtils.clearTable(specTable.activeFruitTypes)
            THUtils.clearTable(specTable.activeFruitTypesArray)
            specTable.functionParameters.isCutterProcessing = false
            specTable.functionParameters.isPickupProcessing = false
            currentConverter = self:thAA_getCutterConverter()
            if currentConverter ~= nil then
                cutterSpec.fruitTypeConverters = currentConverter.data
            end
        end)
    end
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                if workParams.fruitTypesToUse ~= nil then
                    local supportedFillTypes, requiredFillType = getCutterAllowedFillTypes(self)
                    for _, fruitTypeIndex in pairs(workParams.fruitTypesToUse) do
                        local harvestParams = specTable.harvestParameters[fruitTypeIndex]
                        if harvestParams == nil then
                            harvestParams = {}
                            harvestParams.vehicle = self
                            harvestParams.inputFruitType = fruitTypeIndex
                            specTable.harvestParameters[fruitTypeIndex] = harvestParams
                        end
                        harvestParams.supportedFillTypes = supportedFillTypes
                        harvestParams.requiredFillType = requiredFillType
                        harvestParams.isActive = false
                        if not specTable.activeFruitTypes[fruitTypeIndex] then
                            table.insert(specTable.activeFruitTypesArray, fruitTypeIndex)
                            specTable.activeFruitTypes[fruitTypeIndex] = true
                        end
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THConversionCutter:hook_onEndWorkAreaProcessing(superFunc, dt, ...)
    local specTable = getSpecTable(self, false)
    local cutterSpec = self.spec_cutter
    local vehicleName = self:getName()
    local workParams = cutterSpec.workAreaParameters
    local currentConverter = nil
    local isCutterDirty = false
    if specTable ~= nil then
        THUtils.call(function()
            if specTable.fruitTypeConverters.currentIndex > 0 then
                currentConverter, isCutterDirty = self:thAA_setCutterConverter(specTable.fruitTypeConverters.currentIndex, true)
            end
            if workParams.lastRealArea > 0 then
                if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                    THUtils.displayMsg("%s >> Cutter processing information:", vehicleName)
                    THUtils.displayMsg(" - Converter index: %s", specTable.fruitTypeConverters.currentIndex)
                    if currentConverter ~= nil then
                        THUtils.displayMsg(" - Converter name: %s", currentConverter.name)
                        THUtils.displayMsg(" - Converter info:")
                        THUtils.printTable(currentConverter, 2)
                    end
                    if cutterSpec.fruitTypeConverters ~= nil
                        and next(cutterSpec.fruitTypeConverters) ~= nil
                    then
                        THUtils.displayMsg("- Current converters:")
                        for fruitTypeIndex, converterInfo in pairs(cutterSpec.fruitTypeConverters) do
                            local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                            THUtils.displayMsg("[%s]: %s", fruitTypeName, fruitTypeIndex)
                            THUtils.printTable(converterInfo)
                        end
                    end
                end
            end
        end)
    end
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                if currentConverter ~= nil then
                    for fruitTypeIndex, converterInfo in pairs(currentConverter.data) do
                        if currentConverter.lastData[fruitTypeIndex] == nil then
                            currentConverter.lastData[fruitTypeIndex] = {}
                        end
                        for key, value in pairs(converterInfo) do
                            if currentConverter.lastData[fruitTypeIndex][key] ~= value then
                                currentConverter.lastData[fruitTypeIndex][key] = value
                                isCutterDirty = true
                            end
                        end
                    end
                end
                if cutterSpec.currentInputFillType ~= specTable.lastInputFillType then
                    specTable.lastInputFillType = cutterSpec.currentInputFillType
                    isCutterDirty = true
                end
                if self.isServer then
                    cutterSpec.lastAreaBiggerZero = workParams.lastArea > 0
                    if cutterSpec.lastAreaBiggerZero ~= cutterSpec.lastAreaBiggerZeroSent then
                        cutterSpec.lastAreaBiggerZeroSent = cutterSpec.lastAreaBiggerZero
                        self:raiseDirtyFlags(cutterSpec.dirtyFlag)
                        isCutterDirty = true
                    end
                end
                if isCutterDirty then
                    if self.isServer then
                        self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THConversionCutter.inj_cutFruitArea(superFunc, extraArgs, fruitTypeIndex, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and extraArgs.isCutterProcessing
            and fruitTypeIndex ~= nil and specTable.activeFruitTypes[fruitTypeIndex]
        then
            local cutterSpec = self.spec_cutter
            local harvestParams = specTable.harvestParameters[fruitTypeIndex]
            local workParams = cutterSpec.workAreaParameters
            if workParams.combineVehicle ~= nil
                and harvestParams ~= nil and next(harvestParams) ~= nil
            then
                local currentConverter = nil
                THUtils.call(function()
                    currentConverter = self:thAA_getCutterConverter()
                    harvestParams.converterName = THConversionCutter.CONVERTER.UNKNOWN
                    harvestParams.inputFillType = FillType.UNKNOWN
                    harvestParams.modifierIndex = 0
                    harvestParams.isActive = false
                    if currentConverter ~= nil then
                        if self.isServer then
                            currentConverter.modifiers[fruitTypeIndex] = 0
                        end
                        local realConverterInfo = currentConverter.realData[fruitTypeIndex]
                        if realConverterInfo ~= nil then
                            harvestParams.converterName = currentConverter.name
                            harvestParams.inputFillType = realConverterInfo.fillTypeIndex
                            harvestParams.modifierIndex = currentConverter.modifiers[fruitTypeIndex] or 0
                            harvestParams.isActive = true
                        end
                    end
                end)
                local function appendFunc(rChangedArea, rRealArea, ...)
                    THUtils.call(function()
                        if rChangedArea ~= nil and rChangedArea > 0
                            and harvestParams.isActive
                        then
                            if self.isServer then
                                if currentConverter ~= nil then
                                    if harvestParams.modifierIndex ~= currentConverter.modifiers[fruitTypeIndex] then
                                        currentConverter.modifiers[fruitTypeIndex] = harvestParams.modifierIndex
                                        self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
                                    end
                                end
                            end
                            self:thAA_updateCutterConverter(fruitTypeIndex)
                        end
                        harvestParams.isActive = false
                    end)
                    return rChangedArea, rRealArea, ...
                end
                return appendFunc(THDensityMapConversionUtil.xtHook_cutFruitArea(superFunc, harvestParams, fruitTypeIndex, ...))
            end
        end
    end
    return superFunc(fruitTypeIndex, ...)
end
function THConversionCutter.inj_setGroundTypeLayerArea(superFunc, extraArgs, sx, sz, wx, wz, hx, hz, chopperValue, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        local mission = g_thConversionSystem.mission
        if specTable ~= nil and mission ~= nil
            and extraArgs.isCutterProcessing
        then
            local fieldGroundSystem = mission.fieldGroundSystem
            local cutterSpec = self.spec_cutter
            local workParams = cutterSpec.workAreaParameters
            THUtils.call(function()
                if workParams.combineVehicle ~= nil
                    and workParams.lastFruitType ~= nil
                    and specTable.activeFruitTypes[workParams.lastFruitType]
                then
                    local fruitTypeInfo = thFruitTypeManager:getFruitType(workParams.lastFruitType)
                    if fruitTypeInfo ~= nil then
                        local fruitTypeChopperValue = fieldGroundSystem:getChopperTypeValue(fruitTypeInfo.chopperTypeIndex)
                        if fruitTypeChopperValue == chopperValue then
                            local modifierData = self:thAA_getCutterFruitTypeModifier(fruitTypeInfo.index)
                            if modifierData ~= nil then
                                if modifierData.values.toChopperType ~= nil then
                                    chopperValue = modifierData.values.toChopperValue
                                    workParams.lastChopperValue = chopperValue
                                end
                            end
                        end
                    end
                end
            end)
            if chopperValue ~= nil then
                return superFunc(sx, sz, wx, wz, hx, hz, chopperValue, ...)
            end
            return 0, 0
        end
    end
    return superFunc(sx, sz, wx, wz, hx, hz, chopperValue, ...)
end
function THConversionCutter.inj_getWindrowFillTypeIndexByFruitTypeIndex(superFunc, extraArgs, manager, fruitTypeIndex, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and extraArgs.isPickupProcessing
            and specTable.lastPickupFruitType ~= nil and specTable.lastPickupFruitType ~= FruitType.UNKNOWN
            and specTable.lastPickupFillType ~= nil and specTable.lastPickupFillType ~= FillType.UNKNOWN
            and fruitTypeIndex == specTable.lastPickupFruitType
        then
            local cutterSpec = self.spec_cutter
            local workParams = cutterSpec.workAreaParameters
            if workParams ~= nil and workParams.combineVehicle ~= nil then
                local function appendFunc(rWindrowIndex, ...)
                    if rWindrowIndex ~= specTable.lastPickupFillType then
                        THUtils.call(function()
                            if THDebugUtil.getIsEnabled(debugFlagId) then
                                local vehicleName = self:getName()
                                local oldWindrowName = thFillTypeManager:getFillTypeName(rWindrowIndex)
                                local newWindrowName = thFillTypeManager:getFillTypeName(specTable.lastPickupFillType)
                                THUtils.displayMsg("%s >> Windrow fillType [%s] (%s) changed to [%s] (%s)", vehicleName, rWindrowIndex, oldWindrowName, specTable.lastPickupFillType, newWindrowName)
                            end
                            rWindrowIndex = specTable.lastPickupFillType
                        end)
                    end
                    return rWindrowIndex, ...
                end
                return appendFunc(superFunc(manager, fruitTypeIndex, ...))
            end
        end
    end
    return superFunc(manager, fruitTypeIndex, ...)
end
function THConversionCutter.inj_pickupFillTypeAroundLine(superFunc, self, amount, fillTypeIndex, ...)
    if self ~= nil then
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and specTable.functionParameters.isPickupProcessing
            and amount ~= nil and amount < 0
            and fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN
            and fillTypeIndex == specTable.lastPickupFillType
        then
            local cutterSpec = self.spec_cutter
            local workParams = cutterSpec.workAreaParameters
            local fruitTypeIndex = specTable.lastPickupFruitType or FruitType.UNKNOWN
            local harvestParams = specTable.harvestParameters[fruitTypeIndex]
            if workParams ~= nil and workParams.combineVehicle ~= nil
                and harvestParams ~= nil and next(harvestParams) ~= nil
            then
                local currentConverter = nil
                THUtils.call(function()
                    currentConverter = self:thAA_getCutterConverter()
                    harvestParams.converterName = THConversionCutter.CONVERTER.UNKNOWN
                    harvestParams.inputFillType = fillTypeIndex
                    harvestParams.modifierIndex = 0
                    harvestParams.isActive = false
                    if currentConverter ~= nil then
                        if self.isServer then
                            currentConverter.modifiers[fruitTypeIndex] = 0
                        end
                        harvestParams.converterName = currentConverter.name
                        harvestParams.modifierIndex = currentConverter.modifiers[fruitTypeIndex] or 0
                        harvestParams.isActive = true
                    end
                end)
                local function appendFunc(rDropped, ...)
                    THUtils.call(function()
                        if rDropped ~= nil and rDropped < 0
                            and harvestParams.isActive
                        then
                            local pickupFillType = harvestParams.inputFillType or fillTypeIndex
                            if self.isServer then
                                if currentConverter ~= nil then
                                    if harvestParams.modifierIndex ~= currentConverter.modifiers[fruitTypeIndex] then
                                        currentConverter.modifiers[fruitTypeIndex] = harvestParams.modifierIndex
                                        self:raiseDirtyFlags(cutterSpec.effectDirtyFlag)
                                    end
                                end
                            end
                            self:thAA_updateCutterConverter(fruitTypeIndex, pickupFillType)
                        end
                        harvestParams.isActive = false
                    end)
                    return rDropped, ...
                end
                return appendFunc(THDensityMapHeightConversionUtil.xtHook_pickupTipToGroundAroundLine(superFunc, harvestParams, self, amount, fillTypeIndex, ...))
            end
        end
    end
    return superFunc(self, amount, fillTypeIndex, ...)
end
g_thConversionSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Cutter", "onStartWorkAreaProcessing", THConversionCutter)
    THUtils.setFunctionHook("Cutter", "onEndWorkAreaProcessing", THConversionCutter)
    THUtils.setFunctionHook("DensityMapHeightUtil", "tipToGroundAroundLine", nil, THConversionCutter.inj_pickupFillTypeAroundLine, false)
end)