-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thConversionSystem.modName
local thFruitTypeManager = g_thConversionSystem.thFruitTypeManager
local thFillTypeManager = g_thConversionSystem.thFillTypeManager
THConversionMower = {}
local debugFlagId = THDebugUtil.createDebugFlag("THConversionMower")
THConversionMower.debugFlagId = debugFlagId
THConversionMower.MOD_NAME = thModName
THConversionMower.SPEC_NAME = "conversionMower"
THConversionMower.CONVERTER = {
    UNKNOWN = "UNKNOWN",
    NONE = "NONE",
    MOWER = "MOWER",
    SWATHER = "SWATHER"
}
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THConversionMower.SPEC_NAME, thModName, verbose)
end
THConversionMower.getSpecTable = getSpecTable
function THConversionMower.initSpecialization()
end
function THConversionMower.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THVehicle, specializations)
        and SpecializationUtil.hasSpecialization(Mower, specializations)
end
function THConversionMower.registerEvents(vehicleType)
end
function THConversionMower.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THConversionMower)
    SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", THConversionMower)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", THConversionMower)
end
function THConversionMower.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "thAA_readConversionMowerFromStream", THConversionMower.thAA_readConversionMowerFromStream)
    SpecializationUtil.registerFunction(vehicleType, "thAA_writeConversionMowerToStream", THConversionMower.thAA_writeConversionMowerToStream)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getMowerConverter", THConversionMower.thAA_getMowerConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_setMowerConverter", THConversionMower.thAA_setMowerConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_addMowerConverter", THConversionMower.thAA_addMowerConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_updateMowerConverter", THConversionMower.thAA_updateMowerConverter)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getMowerFruitTypeModifier", THConversionMower.thAA_getMowerFruitTypeModifier)
end
function THConversionMower.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processMowerArea", THConversionMower.processMowerArea)
end
local function getMowerAllowedFillTypes(self)
    local mowerSpec = self.spec_mower
    local fillUnitSpec = self.spec_fillUnit
    local supportedFillTypes = nil
    local requiredFillType = nil
    if mowerSpec ~= nil and fillUnitSpec ~= nil
        and mowerSpec.fillUnitIndex ~= nil
    then
        local fillUnitIndex = mowerSpec.fillUnitIndex
        local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex)
        if fillUnitFillLevel ~= nil and fillUnitFillLevel > 0 then
            local changeThreshold = self:getFillTypeChangeThreshold(fillUnitIndex)
            if changeThreshold ~= nil and fillUnitFillLevel >= changeThreshold then
                requiredFillType = self:getFillUnitFillType(fillUnitIndex)
            end
        end
        supportedFillTypes = self:getFillUnitSupportedFillTypes(fillUnitIndex)
    end
    return supportedFillTypes, requiredFillType
end
function THConversionMower:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.fruitTypeConverters = {
        byId = {},
        byIndex = {},
        defaultName = THConversionMower.CONVERTER.UNKNOWN,
        defaultIndex = 0,
        currentIndex = 0,
        lastIndex = 0
    }
    specTable.harvestParameters = {}
    specTable.activeFruitTypes = {}
    specTable.activeFruitTypesArray = {}
    specTable.functionParameters = {
        vehicle = self,
        isMowerProcessing = false
    }
    if not specTable.thIsInitialized then
        THUtils.setFunctionHook("FSDensityMapUtil", "cutFruitArea", nil, THConversionMower.inj_cutFruitArea, false, specTable.functionParameters)
        specTable.thIsInitialized = true
    end
end
function THConversionMower:onPostLoad(savegame)
    local specTable = getSpecTable(self)
    local mowerSpec = self.spec_mower
    local vehicleName = self:getName()
    local converterName = THUtils.getXMLValue(self.xmlFile, "vehicle.mower", "#fruitTypeConverter")
    local defaultConverter = self:thAA_addMowerConverter(THConversionMower.CONVERTER.NONE)
    local isSwather = self.xmlFile:getBool("vehicle.mower#thIsSwather", false)
    if converterName ~= nil then
        local realConverter = self:thAA_addMowerConverter(converterName, true, isSwather)
        if realConverter ~= nil then
            defaultConverter = realConverter
        end
    end
    if defaultConverter == nil then
        THUtils.stackMsg("Could not create default mower converter")
        converterName = THConversionMower.CONVERTER.UNKNOWN
    else
        converterName = defaultConverter.name
        specTable.fruitTypeConverters.defaultName = converterName
        specTable.fruitTypeConverters.defaultIndex = defaultConverter.index
        self:thAA_setMowerConverter(defaultConverter.index, true)
    end
    if type(mowerSpec.converterList) == "table"
        and #mowerSpec.converterList > 0
    then
        for _, otherConverterName in ipairs(mowerSpec.converterList) do
            self:thAA_addMowerConverter(otherConverterName)
        end
    end
    if THDebugUtil.getIsEnabled(debugFlagId) then
        THUtils.displayMsg("%s >> Mower information:", vehicleName)
        THUtils.displayMsg("- Converter name: %s", converterName)
        THUtils.printTable(specTable, 1)
        if mowerSpec.fruitTypeConverters ~= nil
            and next(mowerSpec.fruitTypeConverters) ~= nil
        then
            THUtils.displayMsg("- Default mower converters:")
            for fruitTypeIndex, converterInfo in pairs(mowerSpec.fruitTypeConverters) do
                local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                THUtils.displayMsg("[%s]: %s", fruitTypeIndex, fruitTypeName)
                THUtils.printTable(converterInfo)
            end
        end
        if #specTable.fruitTypeConverters.byIndex > 0 then
            THUtils.displayMsg("- Extended mower converters:")
            for _, otherConverter in ipairs(specTable.fruitTypeConverters.byIndex) do
                THUtils.printTable(otherConverter)
                for fruitTypeIndex, otherConverterInfo in pairs(otherConverter.data) do
                    local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                    THUtils.displayMsg("[%s]: %s", fruitTypeName, fruitTypeIndex)
                    THUtils.printTable(otherConverterInfo)
                end
            end
        end
    end
end
function THConversionMower:onDelete()
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        specTable.functionParameters.vehicle = nil -- disables code injections
    end
end
function THConversionMower:thAA_getMowerConverter(converterId)
    local specTable = getSpecTable(self)
    if converterId == nil then
        return specTable.fruitTypeConverters.byIndex[specTable.fruitTypeConverters.currentIndex]
    else
        local idType = type(converterId)
        if idType == "number" then
            return specTable.fruitTypeConverters.byIndex[converterId]
        elseif idType == "string" then
            return specTable.fruitTypeConverters.byId[converterId:upper()]
        end
    end
end
function THConversionMower:thAA_setMowerConverter(converterId, noEventSend)
    local specTable = getSpecTable(self)
    local mowerSpec = self.spec_mower
    local vehicleName = self:getName()
    if converterId == nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "converterId", converterId)
        return nil, false
    end
    local newConverter = self:thAA_getMowerConverter(converterId)
    local isConverterDirty = false
    if newConverter == nil then
        THUtils.warningMsg("%s >> Could not find converter (%s)", vehicleName, converterId)
        return nil, false
    end
    if specTable.fruitTypeConverters.currentIndex ~= newConverter.index
        or specTable.fruitTypeConverters.lastIndex ~= newConverter.index
        or mowerSpec.fruitTypeConverters ~= newConverter.data
    then
        if THDebugUtil.getIsEnabled(debugFlagId) then
            local oldConverter = self:thAA_getMowerConverter(specTable.fruitTypeConverters.lastIndex)
            if oldConverter == nil then
                THUtils.displayMsg("%s >> Mower converter switched to %q", vehicleName, newConverter.name)
            elseif oldConverter.index ~= newConverter.index then
                THUtils.displayMsg("%s >> Mower converter switched from %q to %q", vehicleName, oldConverter.name, newConverter.name)
            end
        end
        specTable.fruitTypeConverters.currentIndex = newConverter.index
        specTable.fruitTypeConverters.lastIndex = newConverter.index
        mowerSpec.fruitTypeConverters = newConverter.data
        if self.isServer and noEventSend ~= true then
            self:raiseDirtyFlags(mowerSpec.dirtyFlag)
        end
        isConverterDirty = true
    end
    return newConverter, isConverterDirty
end
function THConversionMower:thAA_addMowerConverter(name, useCurrent, isSwather)
    local specTable = getSpecTable(self)
    local mowerSpec = self.spec_mower
    local vehicleName = self:getName()
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name) then
        local converterId = name:upper()
        local newConverter = self:thAA_getMowerConverter(converterId)
        if newConverter == nil then
            newConverter = {
                name = converterId,
                index = #specTable.fruitTypeConverters.byIndex + 1,
                data = {},
                lastData = {},
                realData = {},
                modifiers = {},
                isSwather = false
            }
            if converterId ~= THConversionMower.CONVERTER.NONE then
                local realConverter = thFruitTypeManager:getFruitTypeConverterData(converterId)
                if useCurrent == true and mowerSpec.fruitTypeConverters ~= nil then
                    realConverter = mowerSpec.fruitTypeConverters
                end
                if realConverter ~= nil then
                    for fruitTypeIndex, realConverterInfo in pairs(realConverter) do
                        newConverter.realData[fruitTypeIndex] = {}
                        newConverter.data[fruitTypeIndex] = {}
                        for key, value in pairs(realConverterInfo) do
                            newConverter.realData[fruitTypeIndex][key] = value
                            newConverter.data[fruitTypeIndex][key] = value
                        end
                    end
                end
            end
            if isSwather == true then
                local swatherConverter = thFruitTypeManager:getFruitTypeConverterData(THConversionMower.CONVERTER.SWATHER)
                if swatherConverter ~= nil then
                    for fruitTypeIndex, swatherConverterInfo in pairs(swatherConverter) do
                        if newConverter.realData[fruitTypeIndex] == nil then
                            newConverter.realData[fruitTypeIndex] = {}
                        end
                        if newConverter.data[fruitTypeIndex] == nil then
                            newConverter.data[fruitTypeIndex] = {}
                        end
                        for key, value in pairs(swatherConverterInfo) do
                            newConverter.realData[fruitTypeIndex][key] = value
                            newConverter.data[fruitTypeIndex][key] = value
                        end
                    end
                    newConverter.isSwather = true
                end
            end
            specTable.fruitTypeConverters.byIndex[newConverter.index] = newConverter
            specTable.fruitTypeConverters.byId[newConverter.name] = newConverter
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THUtils.displayMsg("%s >> Added mower converter %q", vehicleName, newConverter.name)
                THUtils.printTable(newConverter, 2)
            end
        end
        return newConverter
    end
end
function THConversionMower:thAA_updateMowerConverter(inputFruitType, noEventSend)
    local mowerSpec = self.spec_mower
    local vehicleName = self:getName()
    local currentConverter = self:thAA_getMowerConverter()
    if currentConverter ~= nil then
        local converterInfo = currentConverter.data[inputFruitType]
        local realConverterInfo = currentConverter.realData[inputFruitType]
        local isConverterDirty = false
        if realConverterInfo ~= nil then
            if converterInfo == nil then
                converterInfo = {}
                currentConverter.data[inputFruitType] = converterInfo
            end
            for key, value in pairs(realConverterInfo) do
                converterInfo[key] = value
            end
            local modifierIndex = currentConverter.modifiers[inputFruitType]
            local modifierData = g_thConversionSystem:getFruitTypeModifierDataByIndex(inputFruitType, modifierIndex)
            if modifierData ~= nil then
                if modifierData.values.toFillType ~= nil then
                    converterInfo.fillTypeIndex = modifierData.values.toFillType
                    converterInfo.conversionFactor = 1
                    converterInfo.windrowConversionFactor = 1
                end
                if modifierData.values.toFillTypeFactor ~= nil then
                    converterInfo.conversionFactor = modifierData.values.toFillTypeFactor
                    converterInfo.windrowConversionFactor = modifierData.values.toFillTypeFactor
                end
            end
            if currentConverter.lastData[inputFruitType] == nil then
                currentConverter.lastData[inputFruitType] = {}
            end
            for key, value in pairs(converterInfo) do
                if currentConverter.lastData[inputFruitType][key] ~= value then
                    currentConverter.lastData[inputFruitType][key] = value
                    isConverterDirty = true
                end
            end
            if isConverterDirty then
                if self.isServer and noEventSend ~= true then
                    self:raiseDirtyFlags(mowerSpec.dirtyFlag)
                end
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local inputFruitTypeName = thFruitTypeManager:getFruitTypeName(inputFruitType)
                    local inputFillTypeName = thFillTypeManager:getFillTypeName(realConverterInfo.fillTypeIndex)
                    THUtils.displayMsg("%s >> Mower conversion values:", vehicleName)
                    THUtils.displayMsg("- Converter name: %s", currentConverter.name)
                    THUtils.displayMsg("- Input fruitType: [%s] (%s)", inputFruitType, inputFruitTypeName)
                    THUtils.displayMsg("- Input fillType: [%s] (%s)", realConverterInfo.fillTypeIndex, inputFillTypeName)
                    THUtils.printTable(converterInfo)
                    if modifierData ~= nil then
                        THUtils.displayMsg("- Modifier Values:")
                        THUtils.printTable(modifierData.values)
                        THUtils.displayMsg("- Modifier Conditions:")
                        THUtils.printTable(modifierData.conditions, 1)
                    end
                end
            end
            return true, isConverterDirty
        end
    end
    return false, false
end
function THConversionMower:thAA_getMowerFruitTypeModifier(fruitType)
    local currentConverter = self:thAA_getMowerConverter()
    local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
    if currentConverter ~= nil and fruitTypeInfo ~= nil then
        local modifierIndex = currentConverter.modifiers[fruitTypeInfo.index]
        if modifierIndex ~= nil then
            return g_thConversionSystem:getFruitTypeModifierDataByIndex(fruitTypeInfo.index, modifierIndex)
        end
    end
end
function THConversionMower:thAA_readConversionMowerFromStream(streamId, connection)
    local currentConverter = nil
    if connection:getIsServer() then
        local converterIndex = streamReadUIntN(streamId, 4)
        THUtils.call(function()
            if converterIndex > 0 then
                currentConverter = self:thAA_setMowerConverter(converterIndex, true)
            end
        end)
        local numFruitTypes = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
        if numFruitTypes > 0 then
            for _ = 1, numFruitTypes do
                local fruitTypeIndex = streamReadUIntN(streamId, FruitTypeManager.SEND_NUM_BITS)
                local modifierIndex = streamReadUIntN(streamId, THConversionSystem.MODIFIER_SEND_BITS)
                THUtils.call(function()
                    if currentConverter ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN then
                        currentConverter.modifiers[fruitTypeIndex] = modifierIndex
                        self:thAA_updateMowerConverter(fruitTypeIndex, true)
                    end
                end)
            end
        end
    end
end
function THConversionMower:thAA_writeConversionMowerToStream(streamId, connection)
    local specTable = getSpecTable(self)
    local currentConverter = nil
    if not connection:getIsServer() then
        local converterIndex = specTable.fruitTypeConverters.currentIndex or 0
        streamWriteUIntN(streamId, converterIndex, 4)
        THUtils.call(function()
            currentConverter = self:thAA_getMowerConverter(converterIndex)
        end)
        local numFruitTypes = #specTable.activeFruitTypesArray
        streamWriteUIntN(streamId, numFruitTypes, FruitTypeManager.SEND_NUM_BITS)
        if numFruitTypes > 0 then
            for i = 1, numFruitTypes do
                local fruitTypeIndex = specTable.activeFruitTypesArray[i]
                local modifierIndex = 0
                if currentConverter ~= nil then
                    modifierIndex = currentConverter.modifiers[fruitTypeIndex] or 0
                end
                streamWriteUIntN(streamId, fruitTypeIndex, FruitTypeManager.SEND_NUM_BITS)
                streamWriteUIntN(streamId, modifierIndex, THConversionSystem.MODIFIER_SEND_BITS)
            end
        end
    end
end
function THConversionMower:processMowerArea(superFunc, workArea, ...)
    local specTable = getSpecTable(self)
    specTable.functionParameters.isMowerProcessing = true
    local function appendFunc(...)
        specTable.functionParameters.isMowerProcessing = false
        return ...
    end
    return appendFunc(superFunc(self, workArea, ...))
end
function THConversionMower:switchConverter(superFunc, ...)
    local specTable = getSpecTable(self, false)
    local mowerSpec = self.spec_mower
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                local currentConverter = self:thAA_getMowerConverter()
                if self.isServer
                    and mowerSpec.converterIndex ~= nil
                    and mowerSpec.converterIndex > 0
                    and mowerSpec.currentConverter ~= nil
                then
                    local newConverter = self:thAA_getMowerConverter(mowerSpec.currentConverter)
                    if newConverter ~= nil then
                        currentConverter = newConverter
                    end
                end
                if currentConverter ~= nil then
                    self:thAA_setMowerConverter(currentConverter.index)
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, ...))
end
function THConversionMower:hook_onStartWorkAreaProcessing(superFunc, dt, ...)
    local specTable = getSpecTable(self, false)
    local mowerSpec = self.spec_mower
    local currentConverter = nil
    if specTable ~= nil then
        THUtils.call(function()
            THUtils.clearTable(specTable.activeFruitTypes)
            THUtils.clearTable(specTable.activeFruitTypesArray)
            specTable.functionParameters.isMowerProcessing = false
            currentConverter = self:thAA_getMowerConverter()
            if currentConverter ~= nil then
                mowerSpec.fruitTypeConverters = currentConverter.data
            end
        end)
    end
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                if mowerSpec.fruitTypeConverters ~= nil then
                    local supportedFillTypes, requiredFillType = getMowerAllowedFillTypes(self)
                    for fruitTypeIndex in pairs(mowerSpec.fruitTypeConverters) do
                        local harvestParams = specTable.harvestParameters[fruitTypeIndex]
                        if harvestParams == nil then
                            harvestParams = {}
                            harvestParams.vehicle = self
                            harvestParams.inputFruitType = fruitTypeIndex
                            specTable.harvestParameters[fruitTypeIndex] = harvestParams
                        end
                        harvestParams.supportedFillTypes = supportedFillTypes
                        harvestParams.requiredFillType = requiredFillType
                        harvestParams.isActive = false
                        if not specTable.activeFruitTypes[fruitTypeIndex] then
                            table.insert(specTable.activeFruitTypesArray, fruitTypeIndex)
                            specTable.activeFruitTypes[fruitTypeIndex] = true
                        end
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THConversionMower:hook_onEndWorkAreaProcessing(superFunc, dt, ...)
    local specTable = getSpecTable(self, false)
    local mowerSpec = self.spec_mower
    local vehicleName = self:getName()
    local workParams = mowerSpec.workAreaParameters
    local currentConverter = nil
    local isMowerDirty = false
    if specTable ~= nil then
        THUtils.call(function()
            if specTable.fruitTypeConverters.currentIndex > 0 then
                currentConverter, isMowerDirty = self:thAA_setMowerConverter(specTable.fruitTypeConverters.currentIndex, true)
            end
            if workParams.lastChangedArea > 0 then
                if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                    THUtils.displayMsg("%s >> Mower processing information:", vehicleName)
                    THUtils.displayMsg(" - Converter index: %s", specTable.fruitTypeConverters.currentIndex)
                    if currentConverter ~= nil then
                        THUtils.displayMsg(" - Converter name: %s", currentConverter.name)
                        THUtils.displayMsg(" - Converter info:")
                        THUtils.printTable(currentConverter, 2)
                    end
                    if mowerSpec.fruitTypeConverters ~= nil
                        and next(mowerSpec.fruitTypeConverters) ~= nil
                    then
                        THUtils.displayMsg("- Current converters:")
                        for fruitTypeIndex, converterInfo in pairs(mowerSpec.fruitTypeConverters) do
                            local fruitTypeName = thFruitTypeManager:getFruitTypeName(fruitTypeIndex)
                            THUtils.displayMsg("[%s]: %s", fruitTypeName, fruitTypeIndex)
                            THUtils.printTable(converterInfo)
                        end
                    end
                end
            end
        end)
    end
    local function appendFunc(...)
        if specTable ~= nil then
            THUtils.call(function()
                if currentConverter ~= nil then
                    for fruitTypeIndex, converterInfo in pairs(currentConverter.data) do
                        if currentConverter.lastData[fruitTypeIndex] == nil then
                            currentConverter.lastData[fruitTypeIndex] = {}
                        end
                        for key, value in pairs(converterInfo) do
                            if currentConverter.lastData[fruitTypeIndex][key] ~= value then
                                currentConverter.lastData[fruitTypeIndex][key] = value
                                isMowerDirty = true
                            end
                        end
                    end
                end
                if isMowerDirty then
                    if self.isServer then
                        self:raiseDirtyFlags(mowerSpec.dirtyFlag)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THConversionMower:hook_onReadStream(superFunc, streamId, connection, ...)
    if connection:getIsServer() then
        self:thAA_readConversionMowerFromStream(streamId, connection)
    end
    return superFunc(self, streamId, connection, ...)
end
function THConversionMower:hook_onWriteStream(superFunc, streamId, connection, ...)
    if not connection:getIsServer() then
        self:thAA_writeConversionMowerToStream(streamId, connection)
    end
    return superFunc(self, streamId, connection, ...)
end
function THConversionMower:hook_onReadUpdateStream(superFunc, streamId, timeStamp, connection, ...)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isDirty
            self:thAA_readConversionMowerFromStream(streamId, connection)
        end
    end
    return superFunc(self, streamId, timeStamp, connection, ...)
end
function THConversionMower:hook_onWriteUpdateStream(superFunc, streamId, connection, dirtyMask, ...)
    local mowerSpec = self.spec_mower
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, mowerSpec.dirtyFlag) ~= 0) then
            self:thAA_writeConversionMowerToStream(streamId, connection)
        end
    end
    return superFunc(self, streamId, connection, dirtyMask, ...)
end
function THConversionMower.inj_cutFruitArea(superFunc, extraArgs, fruitTypeIndex, sx, sz, wx, wz, hx, hz, destroySpray, useMinForageState, excludedSprayType, ...)
    if extraArgs.vehicle ~= nil then
        local self = extraArgs.vehicle
        local specTable = getSpecTable(self, false)
        if specTable ~= nil and extraArgs.isMowerProcessing
            and fruitTypeIndex ~= nil and specTable.activeFruitTypes[fruitTypeIndex]
        then
            local mowerSpec = self.spec_mower
            local harvestParams = specTable.harvestParameters[fruitTypeIndex]
            if harvestParams ~= nil and next(harvestParams) ~= nil then
                local currentConverter = nil
                THUtils.call(function()
                    currentConverter = self:thAA_getMowerConverter()
                    harvestParams.converterName = THConversionMower.CONVERTER.UNKNOWN
                    harvestParams.inputFillType = FillType.UNKNOWN
                    harvestParams.modifierIndex = 0
                    harvestParams.isActive = false
                    if currentConverter ~= nil then
                        if self.isServer then
                            currentConverter.modifiers[fruitTypeIndex] = 0
                        end
                        local realConverterInfo = currentConverter.realData[fruitTypeIndex]
                        if realConverterInfo ~= nil then
                            harvestParams.converterName = currentConverter.name
                            harvestParams.inputFillType = realConverterInfo.fillTypeIndex
                            harvestParams.modifierIndex = currentConverter.modifiers[fruitTypeIndex] or 0
                            harvestParams.isActive = true
                        end
                    end
                end)
                local function appendFunc(rChangedArea, rRealArea, ...)
                    THUtils.call(function()
                        if rChangedArea ~= nil and rChangedArea > 0
                            and harvestParams.isActive
                        then
                            if self.isServer then
                                if currentConverter ~= nil then
                                    if harvestParams.modifierIndex ~= currentConverter.modifiers[fruitTypeIndex] then
                                        currentConverter.modifiers[fruitTypeIndex] = harvestParams.modifierIndex
                                        self:raiseDirtyFlags(mowerSpec.dirtyFlag)
                                    end
                                end
                            end
                            self:thAA_updateMowerConverter(fruitTypeIndex)
                        end
                        harvestParams.isActive = false
                    end)
                    return rChangedArea, rRealArea, ...
                end
                return appendFunc(THDensityMapConversionUtil.xtHook_cutFruitArea(superFunc, harvestParams, fruitTypeIndex, sx, sz, wx, wz, hx, hz, destroySpray, useMinForageState, excludedSprayType, ...))
            end
        end
    end
    return superFunc(fruitTypeIndex, sx, sz, wx, wz, hx, hz, destroySpray, useMinForageState, excludedSprayType, ...)
end
g_thConversionSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Mower", "onStartWorkAreaProcessing", THConversionMower)
    THUtils.setFunctionHook("Mower", "onEndWorkAreaProcessing", THConversionMower)
    THUtils.setFunctionHook("Mower", "onReadStream", THConversionMower)
    THUtils.setFunctionHook("Mower", "onWriteStream", THConversionMower)
    THUtils.setFunctionHook("Mower", "onReadUpdateStream", THConversionMower)
    THUtils.setFunctionHook("Mower", "onWriteUpdateStream", THConversionMower)
end)
THUtils.call(function()
    if Mower.switchConverter ~= nil then
        THUtils.setFunctionHook(Mower, "switchConverter", nil, THConversionMower.switchConverter, true)
    end
end)